"""
The constructor class contains all the method, and logic, needed to get the
enumeration, generate objects, and sample objects.
"""
import abc
from collections import Counter
from functools import partial
from typing import Callable, Dict, Generic, Iterator, List, Optional, Set, Tuple

import sympy

from comb_spec_searcher.typing import (
    CombinatorialClassType,
    CombinatorialObjectType,
    Parameters,
    ParametersMap,
    RelianceProfile,
    SubObjects,
    SubRecs,
    SubSamplers,
    SubTerms,
    Terms,
)


class Constructor(abc.ABC, Generic[CombinatorialClassType, CombinatorialObjectType]):
    """The constructor is akin to the 'counting function' in the comb exp paper."""

    @abc.abstractmethod
    def get_equation(
        self, lhs_func: sympy.Function, rhs_funcs: Tuple[sympy.Function, ...]
    ) -> sympy.Eq:
        """
        Return the sympy.Eq in the form lhs_func = f(rhs_funcs).
        """

    @abc.abstractmethod
    def reliance_profile(self, n: int, **parameters: int) -> RelianceProfile:
        """
        Return the reliance profile. That is for the parameters given,
        which parameters of each individual subclass are required.
        """

    @abc.abstractmethod
    def get_terms(
        self, parent_terms: Callable[[int], Terms], subterms: SubTerms, n: int
    ) -> Terms:
        """
        Return the terms for n given the subterms of the children.
        """

    @abc.abstractmethod
    def get_sub_objects(
        self, subobjs: SubObjects, n: int
    ) -> Iterator[
        Tuple[Parameters, Tuple[List[Optional[CombinatorialObjectType]], ...]]
    ]:
        """
        Iterate over all the possibe subobjs/image of the bijection implied by
        the constructor together with the parameters.
        """

    @abc.abstractmethod
    def random_sample_sub_objects(
        self,
        parent_count: int,
        subsamplers: SubSamplers,
        subrecs: SubRecs,
        n: int,
        **parameters: int,
    ) -> Tuple[Optional[CombinatorialObjectType], ...]:
        """Return a randomly sampled subobjs/image of the bijection implied
        by the constructor."""

    @staticmethod
    def param_map(
        child_pos_to_parent_pos: Tuple[Tuple[int, ...], ...],
        num_parent_params: int,
        param: Parameters,
    ) -> Parameters:
        new_params = [0 for _ in range(num_parent_params)]
        for pos, value in enumerate(param):
            parent_pos = child_pos_to_parent_pos[pos]
            for p in parent_pos:
                new_params[p] += value
        return tuple(new_params)

    @staticmethod
    def build_param_map(
        child_pos_to_parent_pos: Tuple[Tuple[int, ...], ...], num_parent_params: int
    ) -> ParametersMap:
        """
        Return a parameters map according to the given pos map.
        """
        return partial(
            Constructor.param_map,
            child_pos_to_parent_pos,
            num_parent_params,
        )

    @abc.abstractmethod
    def equiv(
        self, other: "Constructor", data: Optional[object] = None
    ) -> Tuple[bool, Optional[object]]:
        """Two constructors are equiv if A <self> B and C <other> D are isomorphic for
        any combinatorial classes A, B, C and D where A and C are isomorphic and B and
        D. This must be implemented for bijections. The data argument is to optionally
        pass additional information to the equiv function. The second returned value is
        optional data for bijections. It is used to pass additional arguments to
        determine index ordering. It needs to be JSON compatible for bijection's
        `to_jsonable` and `from_dict` to work.
        """

    @staticmethod
    def extra_params_equiv(
        params1: Tuple[Dict[str, str], ...],
        params2: Tuple[Dict[str, str], ...],
    ) -> bool:
        """A base equiv for extra params that will suffice for most constructors."""
        non_empty1 = tuple(par for par in params1 if par)
        non_empty2 = tuple(par for par in params2 if par)
        n = len(non_empty1)
        if len(non_empty2) != n:
            return False
        if n == 0:
            return True
        return Constructor._extra_params_match_bijection(n, non_empty1, non_empty2)

    @staticmethod
    def _extra_params_match_single(par1: Dict[str, str], par2: Dict[str, str]) -> bool:
        """Two parameter dictionaries are equivalent if they have the same number
        of keys if there is a value with k keys mapping to it, then there is a
        corresponding value with k keys mapping to it in the other one.
        """
        return len(par1) == len(par2) and sorted(
            Counter(par1.values()).values()
        ) == sorted(Counter(par2.values()).values())

    @staticmethod
    def _extra_params_match_bijection(
        n: int, p1: Tuple[Dict[str, str], ...], p2: Tuple[Dict[str, str], ...]
    ) -> bool:
        """Check if there is a bijection between the two tuples of params
        such that they are pairwise equivalent."""

        # Shared variables in recursion
        bijection: List[int] = []
        in_use: Set[int] = set()
        cache: Dict[Tuple[int, int], bool] = {}

        def _backtracking() -> bool:
            for i in range(n):
                if i in in_use:
                    continue
                bijection.append(i)
                in_use.add(i)

                k = len(bijection)
                key = (k - 1, i)
                res = cache.get(key, None)
                if res is None:
                    res = Constructor._extra_params_match_single(p1[key[0]], p2[key[1]])
                    cache[key] = res

                # Check if consistent and either done or recursive success.
                res = res and (k == n or _backtracking())

                bijection.pop()
                in_use.remove(i)
                if res:
                    return True
            return False

        return _backtracking()



from .base import Constructor
from .cartesian import CartesianProduct, Quotient
from .disjoint import Complement, DisjointUnion

__all__ = [
    "Constructor",
    "DisjointUnion",
    "Complement",
    "CartesianProduct",
    "Quotient",
]



"""
A database to search for tree.

The database do not store the strategy to save memory.
"""
import itertools
from typing import (
    TYPE_CHECKING,
    Iterable,
    Iterator,
    MutableMapping,
    Optional,
    Set,
    Tuple,
    Union,
    cast,
)

from comb_spec_searcher.class_db import ClassDB
from comb_spec_searcher.exception import StrategyDoesNotApply
from comb_spec_searcher.strategies.rule import AbstractRule
from comb_spec_searcher.strategies.strategy import AbstractStrategy, StrategyFactory
from comb_spec_searcher.strategies.strategy_pack import StrategyPack
from comb_spec_searcher.typing import RuleKey

from .base import RuleDBBase

if TYPE_CHECKING:
    from comb_spec_searcher import CombinatorialSpecificationSearcher

__all__ = ["RuleDBForgetStrategy"]


# pylint: disable=too-many-ancestors
class RecomputingDict(MutableMapping[RuleKey, AbstractStrategy]):
    """
    A mapping from rules to strategies that recompute the strategy every time it's
    needed instead of storing it in order to save memory.

    If only equiv is set to true the returned strategy will only create two way rules.

    Also in order to save memory we store flat version of the rules, i.e. for a rule
    (a, (b, c,...)) we store (a, b, c,...)
    """

    def __init__(
        self,
        only_equiv: bool,
    ) -> None:
        self._classdb: Optional[ClassDB] = None
        self._pack: Optional[StrategyPack] = None
        self.rules: Set[Tuple[int, ...]] = set()
        self.only_equiv: bool = only_equiv

    def link_searcher(self, classdb: ClassDB, strat_pack: StrategyPack) -> None:
        if self._classdb is not None or self._pack is not None:
            raise RuntimeError("Searcher is alreay linked")
        self._classdb = classdb
        self._pack = strat_pack

    @property
    def classdb(self) -> ClassDB:
        if self._classdb is None:
            raise RuntimeError("Not linked with classdb")
        return self._classdb

    @property
    def pack(self) -> StrategyPack:
        if self._pack is None:
            raise RuntimeError("Not linked with pack")
        return self._pack

    @staticmethod
    def _flatten(tuple_: RuleKey) -> Tuple[int, ...]:
        return (tuple_[0],) + tuple_[1]

    @staticmethod
    def _unflatten(tuple_: Tuple[int, ...]) -> RuleKey:
        return (tuple_[0], tuple_[1:])

    def __getitem__(self, key: RuleKey) -> AbstractStrategy:
        if self._flatten(key) not in self.rules:
            raise KeyError(key)
        possible_labels = (key[0],) + key[1]
        for label, strat in itertools.product(possible_labels, self.pack):
            comb_class = self.classdb.get_class(label)
            if isinstance(strat, StrategyFactory):
                strats_or_rules: Iterable[
                    Union[AbstractRule, AbstractStrategy]
                ] = strat(comb_class)
            else:
                strats_or_rules = [strat]
            for x in strats_or_rules:
                if isinstance(x, AbstractStrategy):
                    try:
                        rule = x(comb_class)
                    except StrategyDoesNotApply:
                        continue
                else:
                    rule = x
                try:
                    start_label = self.classdb.get_label(rule.comb_class)
                    nonempty_children = tuple(
                        c for c in rule.children if not self.classdb.is_empty(c)
                    )
                    end_labels = tuple(
                        sorted(map(self.classdb.get_label, nonempty_children))
                    )
                    if (start_label, end_labels) == key:
                        if self.only_equiv and not rule.is_two_way():
                            continue
                        return rule.strategy
                except StrategyDoesNotApply:
                    pass
        err_message = (
            f"Could not recompute the strategy for the rule {key} with "
            " any of the strategies. Classes are:\n"
        )
        for label in itertools.chain([key[0]], key[1]):
            err_message += f"Label: {label}\n"
            err_message += str(self.classdb.get_class(label)) + "\n"
        raise RuntimeError(err_message)

    def __setitem__(self, key: RuleKey, value: AbstractStrategy) -> None:
        assert not self.only_equiv or len(key[1]) == 1
        self.rules.add(self._flatten(key))

    def __delitem__(self, key: RuleKey) -> None:
        self.rules.remove(self._flatten(key))

    def __iter__(self) -> Iterator[RuleKey]:
        for rule in self.rules:
            yield self._unflatten(rule)

    def __len__(self) -> int:
        return len(self.rules)

    def __contains__(self, key: object) -> bool:
        return self._flatten(cast(RuleKey, key)) in self.rules


class RuleDBForgetStrategy(RuleDBBase):
    def __init__(self) -> None:
        super().__init__()
        self._rule_to_strategy = RecomputingDict(only_equiv=False)
        self._eqv_rule_to_strategy = RecomputingDict(only_equiv=True)

    def link_searcher(self, searcher: "CombinatorialSpecificationSearcher") -> None:
        super().link_searcher(searcher)
        classdb = searcher.classdb
        strat_pack = searcher.strategy_pack
        self.rule_to_strategy.link_searcher(classdb, strat_pack)
        self.eqv_rule_to_strategy.link_searcher(classdb, strat_pack)

    @property
    def rule_to_strategy(self) -> RecomputingDict:
        return self._rule_to_strategy

    @property
    def eqv_rule_to_strategy(self) -> RecomputingDict:
        return self._eqv_rule_to_strategy



from itertools import product
from typing import (
    TYPE_CHECKING,
    Any,
    Dict,
    Generic,
    Iterator,
    List,
    Optional,
    Set,
    Tuple,
    TypeVar,
)

from .combinatorial_class import CombinatorialClass
from .strategies.rule import AbstractRule, Rule

if TYPE_CHECKING:
    from .combinatorial_class import CombinatorialObject
    from .specification import CombinatorialSpecification


ObjType1 = TypeVar("ObjType1", bound="CombinatorialObject")
ClassType1 = TypeVar("ClassType1", bound="CombinatorialClass")
ClassType2 = TypeVar("ClassType2", bound="CombinatorialClass")
ObjType2 = TypeVar("ObjType2", bound="CombinatorialObject")
OrderMap = Dict[Tuple[ClassType1, ClassType2], List[int]]
IndexDataMap = Optional[Dict[Tuple[ClassType1, ClassType2], object]]


class Isomorphism(Generic[ClassType1, ObjType1, ClassType2, ObjType2]):
    """Isomorphism checker."""

    _INVALID, _UNKNOWN, _VALID = range(-1, 2)

    @classmethod
    def check(
        cls,
        spec1: "CombinatorialSpecification[ClassType1, ObjType1]",
        spec2: "CombinatorialSpecification[ClassType2, ObjType2]",
    ) -> bool:
        """Check if two specs are isomorphic."""
        return cls(spec1, spec2).are_isomorphic()

    def __init__(
        self,
        spec1: "CombinatorialSpecification[ClassType1, ObjType1]",
        spec2: "CombinatorialSpecification[ClassType2, ObjType2]",
    ) -> None:
        # The specs
        self._rules1: Dict[
            ClassType1, AbstractRule[ClassType1, ObjType1]
        ] = spec1.rules_dict
        self._rules2: Dict[
            ClassType2, AbstractRule[ClassType2, ObjType2]
        ] = spec2.rules_dict
        self.root1 = spec1.root
        self.root2 = spec2.root

        # Tracking recursion
        self._ancestors: Set[Tuple[ClassType1, ClassType2]] = set()

        # Tracking matchin permutations
        self._order_map: Dict[Tuple[ClassType1, ClassType2], List[int]] = {}

        # Tracking failed matchings
        self._failed: Set[Tuple[ClassType1, ClassType2]] = set()

        # Tracking additional data
        self._index_data: Dict[Tuple[ClassType1, ClassType2], object] = {}

        # Tracking paths in terms of child indices
        self._path_tracker: List[Tuple[Tuple[int, ...], Tuple[int, ...]]] = []

        # Store results
        self._isomorphic = self._are_isomorphic(spec1.root, spec2.root)

    def get_order_data(
        self,
    ) -> Dict[Tuple[ClassType1, ClassType2], object]:
        """Return any index data for nonbijective matches."""
        return self._index_data

    def are_isomorphic(self) -> bool:
        """Check if the two specs are isomorphic."""
        return self._isomorphic

    def get_order(
        self,
    ) -> OrderMap[ClassType1, ClassType2]:
        """Get order map of corresponding nodes."""
        return self._order_map

    def _are_isomorphic(self, node1: ClassType1, node2: ClassType2) -> bool:
        # Both ends of equivilance paths (len = 1 if none)
        eq_path1, eq_path2 = self._get_eq_descendant(node1, node2)

        # The ones that are not on the LHS of equivalence rules
        curr1, curr2 = eq_path1[-1], eq_path2[-1]

        # Rules from curr1 and curr2 to their respective children
        rule1, rule2 = self._rules1[curr1], self._rules2[curr2]

        # Any empty children are ignored
        non_empty_ind1, non_empty_ind2 = (
            tuple(i for i, c in enumerate(rule1.children) if not c.is_empty()),
            tuple(i for i, c in enumerate(rule2.children) if not c.is_empty()),
        )

        # Check for base cases (invalid = -1, valid = 1)
        is_base_case = self._base_cases(
            eq_path1, rule1, non_empty_ind1, eq_path2, rule2, non_empty_ind2
        )
        if is_base_case:
            return bool(is_base_case + 1)

        # Update ancestors for recursion
        self._ancestors.update(product(eq_path1, eq_path2))

        # The number of nonempty children
        n = len(non_empty_ind1)

        # The permutation of matched children indices
        child_order: List[int] = [-1] * n

        # To avoid repeated index pairs, since outside choices have no effect.
        blacklist: Set[Tuple[int, int]] = set()

        # Iterative stack with elements:
        # (index1, index2, indices of the second rule's children already matched)
        stack = [(0, i, {i}) for i in range(n - 1, -1, -1)]

        # Check all pairings of children. Any pairing hat exhaust all nonempty children,
        # such that all are valid will result in current nodes being matched. This is
        # done with backtracking.
        while stack:
            i1, i2, in_use = stack.pop()

            # Already failed
            if (i1, i2) in blacklist:
                continue

            # Track path
            self._path_tracker_push(
                non_empty_ind1[i1],
                non_empty_ind2[i2],
                len(eq_path1) > 1,
                len(eq_path2) > 1,
            )

            # Call _are_isomorphic recursively for current pair
            # and add to blacklist and try again if we fail.
            if not self._are_isomorphic(
                rule1.children[non_empty_ind1[i1]], rule2.children[non_empty_ind2[i2]]
            ):
                blacklist.add((i1, i2))
                self._path_tracker_pop()
                continue
            self._path_tracker_pop()

            # Update permutation
            child_order[i2] = i1

            # If we reach the last index we have succeeded
            if i1 == n - 1:
                self._order_map[(curr1, curr2)] = child_order
                # Since we did not conclude this match by recursion we can remove our
                # current ids from the ancestor set since the next ones we check are
                # not descendants of the current ids (or their equivalences).
                self._ancestors.difference_update(product(eq_path1, eq_path2))
                return True

            # If we have indices to expand, that is done here
            Isomorphism._extend_stack(i1, n, in_use, stack)

        # If stack has been exhausted we have failed to match the ids. We remove them
        # from the ancestors set, memorize the failure and remove any data gathered.
        self._ancestors.difference_update(product(eq_path1, eq_path2))
        self._failed.add((curr1, curr2))
        self._index_data.pop((curr1, curr2), None)
        return False

    def _get_eq_descendant(
        self, node1: ClassType1, node2: ClassType2
    ) -> Tuple[List[ClassType1], List[ClassType2]]:
        """Get path to nodes that are not on the LHS of equivalence rules."""
        rule1, rule2 = self._rules1[node1], self._rules2[node2]
        nodes1, nodes2 = [node1], [node2]
        if rule1.is_equivalence():
            nodes1.append(rule1.children[0])
        if rule2.is_equivalence():
            nodes2.append(rule2.children[0])
        return nodes1, nodes2

    def _base_cases(
        self,
        eq_nodes1: List[ClassType1],
        rule1: AbstractRule[ClassType1, ObjType1],
        non_empty_children1: Tuple[int, ...],
        eq_nodes2: List[ClassType2],
        rule2: AbstractRule[ClassType2, ObjType2],
        non_empty_children2: Tuple[int, ...],
    ) -> int:
        # The ones that are not on the LHS of equivalence rules
        curr1, curr2 = eq_nodes1[-1], eq_nodes2[-1]

        # Already matched
        if (curr1, curr2) in self._order_map:
            return Isomorphism._VALID

        # Already failed
        if (curr1, curr2) in self._failed:
            return Isomorphism._INVALID

        # If different number of non-empty children
        if len(non_empty_children1) != len(non_empty_children2):
            return Isomorphism._INVALID

        # Atoms
        if not rule1.children and not rule2.children:
            if curr1.is_atom() and curr2.is_atom():
                if self._atom_match(curr1, curr2, rule1, rule2):
                    return Isomorphism._VALID
            return Isomorphism._INVALID

        # If different type of rules are applied, the trees are not isomorphic
        assert isinstance(rule1, Rule) and isinstance(rule2, Rule)
        if not self._constructor_match(rule1, rule2, curr1, curr2):
            return Isomorphism._INVALID

        # Check for recursive match
        if any((n1, n2) in self._ancestors for n1, n2 in product(eq_nodes1, eq_nodes2)):
            return Isomorphism._VALID

        return Isomorphism._UNKNOWN

    def _path_tracker_push(
        self, c1: int, c2: int, eq_path1: bool, eq_path2: bool
    ) -> None:
        """Store the path being taken."""
        t1 = (0, c1) if eq_path1 else (c1,)
        t2 = (0, c2) if eq_path2 else (c2,)
        self._path_tracker.append((t1, t2))

    def _path_tracker_pop(self) -> None:
        """Remove last stored step in path."""
        self._path_tracker.pop()

    def _atom_match(
        self,
        atom1: ClassType1,
        atom2: ClassType2,
        rule1: AbstractRule[ClassType1, ObjType1],
        rule2: AbstractRule[ClassType2, ObjType2],
    ) -> bool:
        """Returns true if atoms match, false otherwise."""
        # pylint: disable=no-self-use
        sz1 = next(atom1.objects_of_size(atom1.minimum_size_of_object())).size()
        sz2 = next(atom2.objects_of_size(atom2.minimum_size_of_object())).size()
        return sz1 == sz2 and rule1.get_terms(sz1) == rule2.get_terms(sz2)

    def _constructor_match(
        self,
        rule1: Rule[ClassType1, ObjType1],
        rule2: Rule[ClassType2, ObjType2],
        curr1: ClassType1,
        curr2: ClassType2,
    ) -> bool:
        are_eq, data = rule1.constructor.equiv(rule2.constructor)
        if not are_eq:
            return False
        if data is not None:
            self._index_data[(curr1, curr2)] = data
        return True

    @staticmethod
    def _extend_stack(
        i1: int, n: int, in_use: Set[int], stack: List[Tuple[int, int, Set[int]]]
    ) -> None:
        for i in range(n - 1, -1, -1):
            if i in in_use:
                continue
            stack.append((i1 + 1, i, in_use.union({i})))


class ParseTreeMap(Generic[ClassType1, ObjType1, ClassType2, ObjType2]):
    """
    Map the object by creating parse tree for it while mimicking the steps
    in the other specification. Then backward maps are applied in the other
    parse tree to get the object the original will map to.
    """

    @classmethod
    def map(
        cls,
        domain: "CombinatorialSpecification[ClassType1, ObjType1]",
        codomain: "CombinatorialSpecification[ClassType2, ObjType2]",
        get_order: OrderMap[ClassType1, ClassType2],
        index_data: Dict[Tuple[ClassType1, ClassType2], object],
        obj: ObjType1,
    ) -> ObjType2:
        return cls(domain, codomain, get_order, index_data).map_rec(
            obj, domain.root_rule, codomain.root_rule
        )

    def __init__(
        self,
        domain: "CombinatorialSpecification[ClassType1, ObjType1]",
        codomain: "CombinatorialSpecification[ClassType2, ObjType2]",
        get_order: OrderMap[ClassType1, ClassType2],
        index_data: Dict[Tuple[ClassType1, ClassType2], object],
    ) -> None:
        self.domain = domain
        self.codomain = codomain
        self.get_order = get_order
        self.index_data = index_data

    @staticmethod
    def _min_object(rule: AbstractRule[ClassType2, ObjType2]) -> ObjType2:
        """Get min object from rule's parent."""
        obj: ObjType2 = next(
            rule.comb_class.objects_of_size(rule.comb_class.minimum_size_of_object())
        )
        return obj

    def _get_nonempty(
        self, rule: Rule[ClassType1, ObjType1], children: Tuple[Optional[ObjType1], ...]
    ) -> Tuple[Optional[Tuple[ObjType1, AbstractRule[ClassType1, ObjType1]]], ...]:
        """Get the nonempty children for the domain's rule. We gather both the part
        of the mapped object and the class it belongs to."""
        return tuple(
            (child_obj, self.domain.rules_dict[child])
            if child_obj is not None
            else None
            for child, child_obj in zip(rule.children, children)
            if not child.is_empty()
        )

    def map_rec(
        self,
        obj: Optional[ObjType1],
        rule1: AbstractRule[ClassType1, ObjType1],
        rule2: AbstractRule[ClassType2, ObjType2],
    ) -> ObjType2:
        # If atom, we return the minimum object of the corresponding rule.
        if not rule1.children:
            return ParseTreeMap._min_object(rule2)

        assert isinstance(rule1, Rule) and isinstance(rule2, Rule)
        mapped_obj, idx = rule1.indexed_forward_map(obj)

        if rule2.is_equivalence():
            if not rule1.is_equivalence():
                # Backward map the outcome of a recursion when
                # we move one rule further in spec2 only.
                val: ObjType2 = rule2.indexed_backward_map(
                    (
                        self.map_rec(
                            obj, rule1, self.codomain.rules_dict[rule2.children[0]]
                        ),
                    ),
                    idx,
                )
                return val
            # When both are eq-rules, we treat them normally and set matching order
            order: List[int] = [0]
        elif rule1.is_equivalence():
            # If only domain's rule is eq-rule then we don't backward map for the
            # spec2's rule here but move along with the mapped object and next rule
            # and return the outcome directly.
            return self.map_rec(
                mapped_obj[0], self.domain.rules_dict[rule1.children[0]], rule2
            )
        else:
            # Fetch the matching order to know what parts to recurse together.
            order = self.get_order[(rule1.comb_class, rule2.comb_class)]

        # Sort rule1's children to match those of rule 2
        _children = self._get_nonempty(rule1, mapped_obj)
        child_it: Iterator[
            Optional[Tuple[ObjType1, AbstractRule[ClassType1, ObjType1]]]
        ] = (_children[idx] for idx in order)

        # Recurse, when rule2's is not empty and the mapped object corresponding to
        # the rule1's nonempty class is not None and gather the result into a tuple
        # to backward map with rule2.
        val = rule2.indexed_backward_map(
            tuple(
                self.map_rec(c[0][0], c[0][1], c[1])
                if c is not None and c[0] is not None
                else None
                for c in (
                    None
                    if child.is_empty()
                    else (next(child_it), self.codomain.rules_dict[child])
                    for child in rule2.children
                )
            ),
            idx,
            self.index_data.get((rule1.comb_class, rule2.comb_class)),
        )
        return val


class Bijection(Generic[ClassType1, ObjType1, ClassType2, ObjType2]):
    """A bijection object that contains a map between two specifications."""

    @classmethod
    def construct(
        cls,
        spec: "CombinatorialSpecification[ClassType1, ObjType1]",
        other: "CombinatorialSpecification[ClassType2, ObjType2]",
    ) -> Optional["Bijection[ClassType1, ObjType1, ClassType2, ObjType2]"]:
        """Create a bijection object between two specifications if possible."""
        iso = Isomorphism[ClassType1, ObjType1, ClassType2, ObjType2](spec, other)
        if not iso.are_isomorphic():
            return None
        return cls(spec, other, iso.get_order(), iso.get_order_data())

    def __init__(
        self,
        spec: "CombinatorialSpecification[ClassType1, ObjType1]",
        other: "CombinatorialSpecification[ClassType2, ObjType2]",
        get_order: OrderMap[ClassType1, ClassType2],
        index_data: IndexDataMap = None,
    ):
        self._index_data = {} if index_data is None else index_data
        self._inv_index_data = {
            (t2, t1): data for (t1, t2), data in self._index_data.items()
        }
        self._spec = spec
        self._other = other
        self._get_order = get_order
        self._get_inverse_order = {
            (c2, c1): Bijection._perm_inv(lis) for ((c1, c2), lis) in get_order.items()
        }

    @property
    def domain(self) -> "CombinatorialSpecification[ClassType1, ObjType1]":
        """The root class of the first specification."""
        return self._spec

    @property
    def codomain(self) -> "CombinatorialSpecification[ClassType2, ObjType2]":
        """The root class of the second specification."""
        return self._other

    @staticmethod
    def _perm_inv(perm: List[int]) -> List[int]:
        inv = [0] * len(perm)
        for i, v in enumerate(perm):
            inv[v] = i
        return inv

    def map(self, obj: ObjType1) -> ObjType2:
        """Map an object of the domain to an object of the codomain."""
        return ParseTreeMap[ClassType1, ObjType1, ClassType2, ObjType2].map(
            self._spec, self._other, self._get_order, self._index_data, obj
        )

    def inverse_map(self, obj: ObjType2) -> ObjType1:
        """Map an object of the codomain to an object of the domain."""
        # Swap types and use inverse versions of tuple maps
        return ParseTreeMap[ClassType2, ObjType2, ClassType1, ObjType1].map(
            self._other, self._spec, self._get_inverse_order, self._inv_index_data, obj
        )

    def to_jsonable(self) -> dict:
        """Return a JSON serializable dictionary for the bijection."""
        id_map, classes = self._classes_to_array()
        reconstructed_order: Dict[str, Dict[str, List[int]]] = {}
        reconstructed_data: Dict[str, Dict[str, object]] = {}
        Bijection._populate_json_map(self._get_order, reconstructed_order, id_map)
        Bijection._populate_json_map(self._index_data, reconstructed_data, id_map)
        return {
            "spec": self._spec.to_jsonable(),
            "other": self._other.to_jsonable(),
            "order": reconstructed_order,
            "index_data": reconstructed_data,
            "classes": classes,
        }

    def _classes_to_array(self) -> Tuple[Dict[CombinatorialClass, str], List[dict]]:
        """Dump classes into an array with indices as keys to simplify maps."""
        id_map: Dict[CombinatorialClass, str] = {}
        classes: List[dict] = []
        for c1, c2 in self._get_order.keys():
            if c1 not in id_map:
                id_map[c1] = f"{len(classes)}"
                classes.append(c1.to_jsonable())
            if c2 not in id_map:
                id_map[c2] = f"{len(classes)}"
                classes.append(c2.to_jsonable())
        return id_map, classes

    @staticmethod
    def _populate_json_map(
        tuple_map: Dict[Tuple[ClassType1, ClassType2], Any],
        json_map: Dict[str, Dict[str, Any]],
        id_map: Dict[CombinatorialClass, str],
    ) -> None:
        """Convert tuple maps into valid json maps."""
        for (c1, c2), value in tuple_map.items():
            i1, i2 = id_map[c1], id_map[c2]
            if i1 not in json_map:
                json_map[i1] = {i2: value}
            else:
                json_map[i1][i2] = value

    @classmethod
    def from_dict(cls, d) -> "Bijection":
        """Return the bijection with the dictionary outputter by
        the 'to_jsonable' method.
        """
        # pylint: disable=import-outside-toplevel
        from comb_spec_searcher import CombinatorialSpecification

        spec1 = CombinatorialSpecification[ClassType1, ObjType1].from_dict(d["spec"])
        spec2 = CombinatorialSpecification[ClassType2, ObjType2].from_dict(d["other"])
        order = [CombinatorialClass.from_dict(c) for c in d["classes"]]
        get_order = {
            (order[int(idx1)], order[int(idx2)]): lis
            for idx1, sub_map in d["order"].items()
            for idx2, lis in sub_map.items()
        }
        index_data = {
            (order[int(idx1)], order[int(idx2)]): data
            for idx1, sub_map in d["index_data"].items()
            for idx2, data in sub_map.items()
        }
        return cls(spec1, spec2, get_order, index_data)  # type: ignore



from collections import defaultdict, deque
from typing import (
    TYPE_CHECKING,
    DefaultDict,
    Dict,
    Generic,
    List,
    Optional,
    Set,
    Tuple,
    TypeVar,
)

from comb_spec_searcher.comb_spec_searcher import CombinatorialSpecificationSearcher
from comb_spec_searcher.exception import NoMoreClassesToExpandError
from comb_spec_searcher.rule_db.base import RuleDB
from comb_spec_searcher.specification import CombinatorialSpecification
from comb_spec_searcher.specification_extrator import (
    EquivalenceRuleExtractor,
    SpecificationRuleExtractor,
)
from comb_spec_searcher.strategies.rule import AbstractRule, Rule
from comb_spec_searcher.tree_searcher import Node, prune
from comb_spec_searcher.typing import (
    CombinatorialClassType,
    CombinatorialObjectType,
    RuleKey,
    Terms,
)

if TYPE_CHECKING:
    from comb_spec_searcher.combinatorial_class import (
        CombinatorialClass,
        CombinatorialObject,
    )


ClassType1 = TypeVar("ClassType1", bound="CombinatorialClass")
ObjType1 = TypeVar("ObjType1", bound="CombinatorialObject")
ClassType2 = TypeVar("ClassType2", bound="CombinatorialClass")
ObjType2 = TypeVar("ObjType2", bound="CombinatorialObject")
NO_CONSTRUCTOR = None.__class__
MatchingInfo = DefaultDict[
    Tuple[int, int], Dict[Tuple[Tuple[int, ...], Tuple[int, ...]], List[int]]
]
MatchingInfoSingle = DefaultDict[int, DefaultDict[int, Set[Tuple[int, ...]]]]
SpecMap = Dict[int, Tuple[int, ...]]
RuleClassification = DefaultDict[
    int,
    DefaultDict[
        int,
        List[
            Tuple[
                Tuple[int, ...],
                AbstractRule[CombinatorialClassType, CombinatorialObjectType],
            ]
        ],
    ],
]
EqPathTracker = DefaultDict[
    Tuple[int, int],
    DefaultDict[
        Tuple[int, int],
        Dict[Tuple[Tuple[int, ...], Tuple[int, ...]], bool],
    ],
]


class ParallelInfo(Generic[CombinatorialClassType, CombinatorialObjectType]):
    """Information from searcher essential to finding bijection with another and then
    to create specifications.
    """

    def __init__(
        self,
        searcher: CombinatorialSpecificationSearcher[CombinatorialClassType],
        additional_levels: int = 0,
    ):
        self.searcher = searcher
        self._expand(additional_levels)
        if not isinstance(self.searcher.ruledb, RuleDB):
            raise RuntimeError("Only searcher supported rule db is `RuleDB`.")
        self.ruledb: RuleDB = self.searcher.ruledb

        # Root eq label and actual class
        self.root_eq_label: int = self.searcher.ruledb.equivdb[
            self.searcher.start_label
        ]
        self.root_class = self.searcher.classdb.get_class(self.searcher.start_label)

        # Atom information for comparing them
        self.atom_map: Dict[int, Tuple[int, Terms]] = {}

        # Used to filter out impossible matchings to avoid unnecessary expansions
        self.eq_label_rules = self._construct_eq_label_rules()

    def _expand(self, additional_levels: int) -> None:
        """Expand until at least one spec exists and after that some
        additional expansions can be done (optionally)."""
        self._expand_until_spec()
        self._additional_expands(additional_levels)

    def _expand_until_spec(self) -> None:
        try:
            while not self.searcher.has_specification():
                self.searcher.do_level()
        except NoMoreClassesToExpandError as ex:
            if not self.searcher.has_specification():
                raise ValueError("No specifications were found") from ex

    def _additional_expands(self, additional_levels: int) -> None:
        try:
            for _ in range(additional_levels):
                self.searcher.do_level()
        except NoMoreClassesToExpandError:
            pass

    def _construct_eq_label_rules(
        self,
    ) -> RuleClassification[CombinatorialClassType, CombinatorialObjectType]:
        """
        Creates a dictionary d such that
            d[eq_label][child_count]
        is a set of possible resulting children eq_labels and the corresponding
        constructor if any.
        """
        lis = self._pruned_rules_up_to_eq()
        rule_dict = self.ruledb.rule_from_equivalence_rule_dict(lis)
        eq_label_rules: RuleClassification[
            CombinatorialClassType, CombinatorialObjectType
        ] = defaultdict(lambda: defaultdict(list))

        for eq_par, eq_chi in lis:
            parent, rule = self._get_class_and_rule(eq_par, eq_chi, rule_dict)
            assert not parent.is_empty()
            if parent.is_atom():
                sz = next(
                    parent.objects_of_size(parent.minimum_size_of_object())
                ).size()
                self.atom_map[eq_par] = (sz, rule.get_terms(sz))
            else:
                if not isinstance(rule, Rule):
                    raise ValueError("Only atoms can be verified.")
                assert len(eq_chi) > 0
            eq_label_rules[eq_par][len(eq_chi)].append((eq_chi, rule))
        return eq_label_rules

    def _get_class_and_rule(
        self, eq_par: int, eq_chi: Tuple[int, ...], rule_dict: Dict[RuleKey, RuleKey]
    ) -> Tuple[
        CombinatorialClassType,
        AbstractRule[CombinatorialClassType, CombinatorialObjectType],
    ]:
        actual_par, actual_children = rule_dict[(eq_par, eq_chi)]
        strategy = self.ruledb.rule_to_strategy[(actual_par, actual_children)]
        parent = self.searcher.classdb.get_class(actual_par)
        rule: AbstractRule[CombinatorialClassType, CombinatorialObjectType] = strategy(
            parent
        )
        return parent, rule

    def _pruned_rules_up_to_eq(
        self,
    ) -> List[Tuple[int, Tuple[int, ...]]]:
        rules_up_to_eq = self.ruledb.rules_up_to_equivalence()
        prune(rules_up_to_eq)
        return [(k, c) for k, v in rules_up_to_eq.items() for c in v]


class ParallelSpecFinder(Generic[ClassType1, ObjType1, ClassType2, ObjType2]):
    """
    Finder for paralell specs. This version assumes that any classes that share
    equivalence labels are in fact equivalent.
    """

    _INVALID, _UNKNOWN, _VALID = range(-1, 2)

    def __init__(
        self,
        searcher1: CombinatorialSpecificationSearcher[ClassType1],
        searcher2: CombinatorialSpecificationSearcher[ClassType2],
    ):
        self._pre_expand(searcher1, searcher2)
        self._pi1 = ParallelInfo[ClassType1, ObjType1](searcher1)
        self._pi2 = ParallelInfo[ClassType2, ObjType2](searcher2)
        self._ancestors: Set[Tuple[int, int]] = set()

    def _pre_expand(
        self,
        searcher1: CombinatorialSpecificationSearcher[ClassType1],
        searcher2: CombinatorialSpecificationSearcher[ClassType2],
    ) -> None:
        """When overwritten, this method can be implemented and will be called
        prior to the searcher expanding the universe."""

    # First search

    def find(
        self,
    ) -> Optional[
        Tuple[
            CombinatorialSpecification[ClassType1, ObjType1],
            CombinatorialSpecification[ClassType2, ObjType2],
        ]
    ]:
        """Find bijections between the two universes."""
        # d[(p1,p2)][(c1,c2)] = permutation of children such that they match
        matching_info: MatchingInfo = defaultdict(dict)

        # Failed matches
        visited: Set[Tuple[int, int]] = set()

        # Find matching eq lables
        found = self._find(
            self._pi1.root_eq_label, self._pi2.root_eq_label, matching_info, visited
        )

        # Try to construct valid specs from said eq-label matchings
        return self._matching_info_to_specs(found, matching_info)

    def _find(
        self,
        id1: int,
        id2: int,
        matching_info: MatchingInfo,
        visited: Set[Tuple[int, int]],
    ) -> bool:
        """Recursion call to find matching for eq labels (id1,id2)."""

        # Check for base cases (invalid = -1, valid = 1)
        known = self._base_case(id1, id2, matching_info, visited)
        if known:
            return bool(known + 1)

        # Store pair as ancestors ids checked with further recursions
        self._ancestors.add((id1, id2))

        for children1, children2, n in self._potential_children(id1, id2):
            # The permutation of matched children indices
            child_order: List[int] = [-1] * n

            # Iterative stack with elements:
            # (index1, index2, indices of the second rule's children already matched)
            stack = [(0, i, {i}) for i in range(n - 1, -1, -1)]

            # To avoid repeated index pairs, since outside choices have no effect.
            blacklist: Set[Tuple[int, int]] = set()

            # Check all pairings of children. Any pairing hat exhaust all nonempty
            # children, such that all are valid will result in current nodes being
            # matched. This is done with backtracking.
            while stack:
                i1, i2, in_use = stack.pop()

                # Already failed
                if (i1, i2) in blacklist:
                    continue

                # Recursion. Blacklist children pair if we fail.
                if not self._find(children1[i1], children2[i2], matching_info, visited):
                    blacklist.add((i1, i2))
                    continue

                # Update permutation
                child_order[i2] = i1

                # If we reach the last index we have succeeded. Gather this matching and
                # keep trying other children as we want to collect all that are valid.
                if i1 == n - 1:
                    matching_info[(id1, id2)][(children1, children2)] = child_order
                    break

                # Extand stack if we are not done.
                ParallelSpecFinder._extend_stack(i1, n, in_use, stack)

        # Remove self as ancestors if we are done.
        self._ancestors.remove((id1, id2))

        # Store pair in memory of failures.
        visited.add((id1, id2))

        # Return true if any valid matching was found.
        return (id1, id2) in matching_info

    def _base_case(
        self,
        id1: int,
        id2: int,
        matching_info: MatchingInfo,
        visited: Set[Tuple[int, int]],
    ) -> int:
        # If no children we compare atoms
        if (
            id1 in self._pi1.atom_map
            and id2 in self._pi2.atom_map
            and self._atom_match(id1, id2)
        ):
            matching_info[(id1, id2)] = {((), ()): []}
            return ParallelSpecFinder._VALID

        # Already matched
        if (id1, id2) in matching_info:
            return ParallelSpecFinder._VALID
        # Already failed to match
        if (id1, id2) in visited:
            return ParallelSpecFinder._INVALID
        # Recursive match
        if (id1, id2) in self._ancestors:
            return ParallelSpecFinder._VALID
        return ParallelSpecFinder._UNKNOWN

    def _atom_match(self, id1: int, id2: int) -> bool:
        """Returns true if atoms match."""
        sz1, terms1 = self._pi1.atom_map[id1]
        sz2, terms2 = self._pi2.atom_map[id2]
        return sz1 == sz2 and terms1 == terms2

    def _rule_match(
        self,
        rule1: AbstractRule[ClassType1, ObjType1],
        rule2: AbstractRule[ClassType2, ObjType2],
    ) -> bool:
        # pylint: disable=no-self-use
        if not isinstance(rule1, Rule):
            return not isinstance(rule2, Rule)
        if not isinstance(rule2, Rule):
            return False
        return rule1.constructor.equiv(rule2.constructor)[0]

    def _potential_children(
        self, id1: int, id2: int
    ) -> List[Tuple[Tuple[int, ...], Tuple[int, ...], int]]:
        """Children are potential children if all of the following holds they have the
        same number of non-empty children and the constructor's GF matches.
        """
        return [
            (c1, c2, n)
            for n, child_rule_pairs in self._pi1.eq_label_rules[id1].items()
            for c1, r1 in child_rule_pairs
            for c2, r2 in self._pi2.eq_label_rules[id2][n]
            if self._rule_match(r1, r2)
        ]

    @staticmethod
    def _extend_stack(
        i1: int, n: int, in_use: Set[int], stack: List[Tuple[int, int, Set[int]]]
    ) -> None:
        for i in range(n - 1, -1, -1):
            if i in in_use:
                continue
            stack.append((i1 + 1, i, in_use.union({i})))

    # Second search

    def _matching_info_to_specs(
        self,
        found: bool,
        matching_info: MatchingInfo,
    ) -> Optional[
        Tuple[
            CombinatorialSpecification[ClassType1, ObjType1],
            CombinatorialSpecification[ClassType2, ObjType2],
        ]
    ]:
        # If first search was unsuccessful
        if not found:
            return None
        # Attempt to construct specifications from matches
        spec_rules = self._search_matching_info(matching_info)
        if spec_rules is None:
            return None
        return (
            ParallelSpecFinder._create_spec(spec_rules[0], self._pi1),
            ParallelSpecFinder._create_spec(spec_rules[1], self._pi2),
        )

    def _search_matching_info(
        self, matching_info: MatchingInfo
    ) -> Optional[Tuple[SpecMap, SpecMap]]:
        # Matching info 1 and 2 are flatten dictionaries of matching_info
        # for the perspective of each spec. The maps sp1 and sp2 are the
        # gathered specifications (in terms of eq labels), initially empty.
        (
            matching_info1,
            matching_info2,
            sp1,
            sp2,
        ) = ParallelSpecFinder._search_matching_info_init(matching_info)

        # Recursive helper function that tries to populate the spec maps in a way
        # that no label appears on the LHS more than one time.
        def _rec(
            id1: int,
            id2: int,
            id_sets: Tuple[Set[int], Set[int]],
        ) -> bool:
            # Check base cases (invalid = -1, valid = 1)
            bc = ParallelSpecFinder._search_matching_info_recursion_base_cases(
                id1, id2, matching_info, matching_info1, matching_info2, sp1, sp2
            )
            if bc:
                return bool(bc + 1)

            # If ids have appeared before, we don't want to clean them when done
            rec1, rec2 = id1 in sp1, id2 in sp2

            # For all children that matches for given parent pair where they are either
            # not in spec maps or they are mapped to the children we are looking at.
            for children1, children2 in filter(
                lambda c: (id1 not in sp1 or c[0] == sp1[id1])
                and (id2 not in sp2 or c[1] == sp2[id2]),
                matching_info[(id1, id2)],
            ):
                # Add rule for spec to (id -> children) for both specs
                sp1[id1], sp2[id2] = children1, children2

                # Construct cleaning sets that are passed down the recursion
                to_clean: Tuple[Set[int], Set[int]] = (set(), set())

                # If all the children are valid, compared one at a time in the order
                # they were matched in the first search.
                if all(
                    _rec(child1, child2, to_clean)
                    for child1, child2 in zip(
                        (
                            children1[i]
                            for i in matching_info[(id1, id2)][(children1, children2)]
                        ),
                        children2,
                    )
                ):
                    # Update the cleaning set so any failing ancestor can remove.
                    id_sets[0].update(to_clean[0], () if rec1 else (id1,))
                    id_sets[1].update(to_clean[1], () if rec2 else (id2,))
                    return True
                # If failed, remove all descendants that populated the spec maps.
                ParallelSpecFinder._clean_descendants(
                    *to_clean, id1, id2, sp1, sp2, rec1, rec2
                )
            return False

        if _rec(
            self._pi1.root_eq_label,
            self._pi2.root_eq_label,
            (set(), set()),
        ):
            return sp1, sp2
        return None

    @staticmethod
    def _search_matching_info_init(
        matching_info: MatchingInfo,
    ) -> Tuple[MatchingInfoSingle, MatchingInfoSingle, SpecMap, SpecMap]:
        matching_info_1: MatchingInfoSingle = defaultdict(lambda: defaultdict(set))
        matching_info_2: MatchingInfoSingle = defaultdict(lambda: defaultdict(set))
        # convert (p1,p2) -> children to p1 -> p2 -> children and p2-> p1 -> children.
        for (p1, p2), children in matching_info.items():
            for ch1, ch2 in children:
                matching_info_1[p1][p2].add(ch1)
                matching_info_2[p2][p1].add(ch2)
        sp1: SpecMap = {}
        sp2: SpecMap = {}
        return matching_info_1, matching_info_2, sp1, sp2

    @staticmethod
    def _search_matching_info_recursion_base_cases(
        id1: int,
        id2: int,
        matching_info: MatchingInfo,
        matching_info1: MatchingInfoSingle,
        matching_info2: MatchingInfoSingle,
        sp1: SpecMap,
        sp2: SpecMap,
    ) -> int:
        if ParallelSpecFinder._inconsistent_with_matching_info(
            id1, id2, matching_info, matching_info1, matching_info2, sp1, sp2
        ):
            return ParallelSpecFinder._INVALID
        # If atoms, there is only one possible outcome
        if ((), ()) in matching_info[(id1, id2)]:
            sp1[id1], sp2[id2] = (), ()
            return ParallelSpecFinder._VALID
        # If both are assigned, we are done
        if id1 in sp1 and id2 in sp2:
            return ParallelSpecFinder._VALID
        return ParallelSpecFinder._UNKNOWN

    @staticmethod
    def _inconsistent_with_matching_info(
        id1: int,
        id2: int,
        matching_info: MatchingInfo,
        matching_info1: MatchingInfoSingle,
        matching_info2: MatchingInfoSingle,
        sp1: SpecMap,
        sp2: SpecMap,
    ):
        """If ids are not in the matching order or if one is assigned but those
        children aren't compatible with the other.
        """
        return (
            (id1, id2) not in matching_info
            or (id1 in sp1 and sp1[id1] not in matching_info1[id1][id2])
            or (id2 in sp2 and sp2[id2] not in matching_info2[id2][id1])
        )

    @staticmethod
    def _clean_descendants(
        to_clean1: Set[int],
        to_clean2: Set[int],
        id1: int,
        id2: int,
        sp1: SpecMap,
        sp2: SpecMap,
        rec1: bool,
        rec2: bool,
    ):
        # Clean descendants
        for i in to_clean1:
            if i in sp1:
                del sp1[i]
        for i in to_clean2:
            if i in sp2:
                del sp2[i]
        # Clean self if no prior assignments
        if not rec1 and id1 in sp1:
            del sp1[id1]
        if not rec2 and id2 in sp2:
            del sp2[id2]

    # specification construction helpers

    @staticmethod
    def _create_spec(d: SpecMap, pi: ParallelInfo) -> CombinatorialSpecification:
        rules = SpecificationRuleExtractor(
            pi.root_eq_label,
            ParallelSpecFinder._create_tree(d, pi.root_eq_label),
            pi.ruledb,
            pi.searcher.classdb,
        ).rules()
        return CombinatorialSpecification(pi.root_class, rules)

    @staticmethod
    def _create_tree(d: SpecMap, root_eq_label: int) -> Node:
        visited: Set[int] = set()
        root_node = Node(root_eq_label)
        queue = deque([root_node])
        # BFS to consume spec map.
        while queue:
            v = queue.popleft()
            rule = d.get(v.label, ())
            if not (v.label in visited or rule == ()):
                children = [Node(i) for i in rule]
                queue.extend(children)
                v.children = children
            visited.add(v.label)
        return root_node


class EqPathParallelSpecFinder(
    ParallelSpecFinder[ClassType1, ObjType1, ClassType2, ObjType2]
):
    """
    A version of ParallelSpecFinder that supports nonequivalent
    classes sharing equivalence labels.
    """

    def __init__(
        self,
        searcher1: CombinatorialSpecificationSearcher[ClassType1],
        searcher2: CombinatorialSpecificationSearcher[ClassType2],
    ):
        super().__init__(searcher1, searcher2)

        # Tracks the path taken in the second search
        self._path: List[Tuple[int, int, int, int]] = [(-1, -1, -1, -1)]
        self._path_ancestors: Set[Tuple[int, int]] = set()

    def _search_matching_info(
        self, matching_info: MatchingInfo
    ) -> Optional[Tuple[SpecMap, SpecMap]]:
        # This will differ (from super's) that equiv lables must
        # be validated given the path that has been taken.

        (
            matching_info1,
            matching_info2,
            sp1,
            sp2,
        ) = EqPathParallelSpecFinder._search_matching_info_init(matching_info)

        # For storing results of eq path checking.
        eq_path_tracker: EqPathTracker = defaultdict(lambda: defaultdict(dict))

        def _rec(
            id1: int,
            id2: int,
            id_sets: Tuple[Set[int], Set[int]],
        ) -> bool:
            bc = self._search_matching_info_recursion_base_cases_eq(
                id1,
                id2,
                matching_info,
                matching_info1,
                matching_info2,
                sp1,
                sp2,
                self._path[-1],
                eq_path_tracker,
            )
            if bc:
                return bool(bc + 1)
            rec1, rec2 = id1 in sp1, id2 in sp2
            for children1, children2 in filter(
                lambda c: (id1 not in sp1 or c[0] == sp1[id1])
                and (id2 not in sp2 or c[1] == sp2[id2]),
                matching_info[(id1, id2)],
            ):
                sp1[id1], sp2[id2] = children1, children2
                to_clean: Tuple[Set[int], Set[int]] = (set(), set())

                # Check if the path required for the eq labels actually matches
                if self._eq_path_matches(
                    id1, id2, *self._path[-1], sp1, sp2, eq_path_tracker
                ):
                    valid = True
                    for j2, ((j1, child1), child2) in enumerate(
                        zip(
                            (
                                (i, children1[i])
                                for i in matching_info[(id1, id2)][
                                    (children1, children2)
                                ]
                            ),
                            children2,
                        )
                    ):
                        self._path.append((id1, id2, j1, j2))
                        self._path_ancestors.add((id1, id2))
                        if not _rec(child1, child2, to_clean):
                            valid = False
                            self._path.pop()
                            self._path_ancestors.remove((id1, id2))
                            break
                        self._path.pop()
                        self._path_ancestors.remove((id1, id2))
                    if valid:
                        id_sets[0].update(to_clean[0], () if rec1 else (id1,))
                        id_sets[1].update(to_clean[1], () if rec2 else (id2,))
                        return True
                EqPathParallelSpecFinder._clean_descendants(
                    *to_clean, id1, id2, sp1, sp2, rec1, rec2
                )
            return False

        if _rec(
            self._pi1.root_eq_label,
            self._pi2.root_eq_label,
            (set(), set()),
        ):
            return sp1, sp2
        return None

    def _search_matching_info_recursion_base_cases_eq(
        self,
        id1: int,
        id2: int,
        matching_info: MatchingInfo,
        matching_info1: MatchingInfoSingle,
        matching_info2: MatchingInfoSingle,
        sp1: SpecMap,
        sp2: SpecMap,
        relations: Tuple[int, int, int, int],
        eq_path_tracker: EqPathTracker,
    ) -> int:
        pid1, pid2, idx1, idx2 = relations
        if EqPathParallelSpecFinder._inconsistent_with_matching_info(
            id1, id2, matching_info, matching_info1, matching_info2, sp1, sp2
        ):
            return EqPathParallelSpecFinder._INVALID
        # These atoms will have matches in the first searcher but we do offer
        # additional checks which will, by default, always return true.
        if ((), ()) in matching_info[(id1, id2)]:
            if not self._atom_path_match(id1, id2, sp1, sp2):
                return EqPathParallelSpecFinder._INVALID
            sp1[id1], sp2[id2] = (), ()
            return EqPathParallelSpecFinder._VALID
        # If both ids are set we still need to check if they are valid in terms
        # of eq paths since we may have arrived from different parents.
        if id1 in sp1 and id2 in sp2:
            if self._eq_path_matches(
                id1, id2, pid1, pid2, idx1, idx2, sp1, sp2, eq_path_tracker
            ):
                # If we have already matched id1 and id2 but in a different path, they
                # can still be incompatible due to this current path's atom labels.
                should_check = (id1, id2) not in self._path_ancestors
                if should_check and not self._validate_atoms_for_existing_entries(
                    id1, id2, sp1, sp2, matching_info, set()
                ):
                    return EqPathParallelSpecFinder._INVALID
                return EqPathParallelSpecFinder._VALID
            return EqPathParallelSpecFinder._INVALID
        return EqPathParallelSpecFinder._UNKNOWN

    def _validate_atoms_for_existing_entries(
        self,
        id1: int,
        id2: int,
        sp1: SpecMap,
        sp2: SpecMap,
        matching_info: MatchingInfo,
        mem: Set[Tuple[int, int]],
    ) -> bool:
        """Check if atoms are valid for a given path, when initial ids have already
        been added to the spec maps elsewhere."""
        if (id1, id2) in mem:
            return True
        children1, children2 = sp1.get(id1, None), sp2.get(id2, None)
        if children1 is None or children2 is None:
            return True
        if children1 == () == children2:
            return self._atom_path_match(id1, id2, sp1, sp2)
        mem.add((id1, id2))
        for j2, ((j1, child1), child2) in enumerate(
            zip(
                (
                    (i, children1[i])
                    for i in matching_info[(id1, id2)][(children1, children2)]
                ),
                children2,
            )
        ):
            self._path.append((id1, id2, j1, j2))
            check_descendant = self._validate_atoms_for_existing_entries(
                child1, child2, sp1, sp2, matching_info, mem
            )
            self._path.pop()
            if not check_descendant:
                return False
        return True

    def _atom_path_match(self, id1: int, id2: int, sp1: SpecMap, sp2: SpecMap) -> bool:
        """This can be overwritten if the path can affect the validity of the pair."""
        # pylint: disable=no-self-use
        return True

    def _eq_path_matches(
        self,
        id1,
        id2,
        pid1,
        pid2,
        idx1: int,
        idx2: int,
        sp1: SpecMap,
        sp2: SpecMap,
        cache: EqPathTracker,
    ) -> bool:
        """
        Given (grandparent, parent, children) eq labels in both specs, this will check
        if the actual rules required (disregarding actual equivalence rules) to traverse
        this path match.
        """
        children = (sp1[id1], sp2[id2])
        children_cache = cache[(id1, id2)][(pid1, pid2)]
        if children not in children_cache:
            path1 = EquivalenceRuleExtractor(
                self._pi1.root_eq_label,
                self._pi1.searcher.start_label,
                EqPathParallelSpecFinder._create_tree(sp1, self._pi1.root_eq_label),
                self._pi1.ruledb,
                self._pi1.searcher.classdb,
                id1,
                pid1,
                idx1,
            ).nonequivalent_rules_in_equiv_path()
            path2 = EquivalenceRuleExtractor(
                self._pi2.root_eq_label,
                self._pi2.searcher.start_label,
                EqPathParallelSpecFinder._create_tree(sp2, self._pi2.root_eq_label),
                self._pi2.ruledb,
                self._pi2.searcher.classdb,
                id2,
                pid2,
                idx2,
            ).nonequivalent_rules_in_equiv_path()
            children_cache[children] = len(path1) == len(path2) and all(
                self._rule_match(r1, r2) for r1, r2 in zip(path1, path2)
            )
        return children_cache[children]



"""
A database to keep track of equivalent combinatorial classes.

This is done using a union find method. Also, explanations of how combinatorial
classes  are equivalent are maintained.

Based on: https://www.ics.uci.edu/~eppstein/PADS/UnionFind.py

In this file a combinatorial class is represented by a label, which is an
integer, that the classdb gives.
"""

from collections import defaultdict, deque
from typing import Deque, Dict, Iterator, List, Set, Tuple

from .utils import cssmethodtimer


class EquivalenceDB:
    """
    A database for equivalences. Supports four methods.

    - DB[x] return a name for the set containing x. Each set named by
    arbitrarily chosen member.
    - DB.union(t1, t2, explanation) merges the sets containing t1 and t2 and
    records why t1 and t2 are equivalent.
    - DB.equivalent(t1, t2) returns True, if t1 and t2 are in the same set,
    otherwise returns False.
    - DB.get_relation(t1, t2) returns a string explaining why t1 and t2 are
      equivalent.
    """

    def __init__(self) -> None:
        """Create a new empty equivalent database."""
        self.parents: Dict[int, int] = {}
        self.weights: Dict[int, int] = {}
        self.verified_roots: Set[int] = set()
        self.vertices: Dict[int, Set[int]] = defaultdict(set)
        self._one_way_vertices: Dict[int, Set[int]] = defaultdict(set)
        self.func_times: Dict[str, float] = defaultdict(float)
        self.func_calls: Dict[str, int] = defaultdict(int)

    def __eq__(self, other: object) -> bool:
        """Check if all information stored is the same."""
        if not isinstance(other, EquivalenceDB):
            return NotImplemented
        return bool(
            self.parents == other.parents
            and self.weights == other.weights
            and self.verified_roots == other.verified_roots
            and self.vertices == other.vertices
        )

    def __getitem__(self, comb_class: int) -> int:
        """Find and return root combinatorial class for the set containing
        comb_class."""
        root = self.parents.get(comb_class)
        if root is None:
            self.parents[comb_class] = comb_class
            self.weights[comb_class] = 1
            return comb_class

        # Find path of combinatorial class leading to root.
        path = [comb_class]
        while root != path[-1]:
            path.append(root)
            root = self.parents[root]

        # Compress the path and return
        for ancestor in path:
            self.parents[ancestor] = root
        return root

    def get_one_way_vertices(self) -> Dict[int, Set[int]]:
        """
        Return the adjacency table where vertices are strong components,
        and edges are those that connect some vertex in a strong component
        to the other.

        Note: this might not be accurate, you should use the connect_cycle
        method first if needed.
        """
        res: Dict[int, Set[int]] = defaultdict(set)
        for start, ends in self._one_way_vertices.items():
            start = self[start]
            for end in ends:
                end = self[end]
                if start != end:
                    res[start].add(end)
        self._one_way_vertices = res
        return res

    def add_one_way_edge(self, label: int, other_label: int) -> None:
        """Add an edge from label to other label in the directed graph."""
        self._add_edge(label, other_label)
        self._one_way_vertices[self[label]].add(self[other_label])

    @cssmethodtimer("find_paths")
    def connect_cycles(self):
        """Look for cycles using one way edges that have been added."""
        one_way_vertices = self.get_one_way_vertices()
        stack: List[Tuple[int, ...]] = []
        stack.extend([(label,) for label in one_way_vertices.keys()])
        visited = set()
        while stack:
            path = stack.pop()
            end = path[-1]
            if end in visited:
                continue
            visited.add(end)
            for new_end in one_way_vertices[end]:
                for i, vertex in enumerate(path[:-1]):
                    if self.equivalent(vertex, new_end):
                        # we've detected a cycle!
                        for eqv_vertix in path[i:]:
                            self._set_equivalent(eqv_vertix, new_end)
                        break
                if new_end not in path:
                    # new end is not in the path, so we haven't tried looking for
                    # cycles from here.
                    stack.append(
                        path + (new_end,)
                    )  # appending left makes this a depth first search

    def add_two_way_edge(self, label: int, other_label: int) -> None:
        """Add a two way edge to the directed graph."""
        self._add_edge(label, other_label)
        self._add_edge(other_label, label)
        self._set_equivalent(label, other_label)

    def _add_edge(self, label: int, other_label: int) -> None:
        """Add an edge from label to other_label."""
        if label == other_label:
            return
        self.vertices[label].add(other_label)

    def find_paths(self, label: int, other_label: int) -> Iterator[Tuple[int, ...]]:
        """Return an iterator of paths starting at other_label and ending at label."""
        dequeue: Deque[Tuple[int, ...]] = deque()
        dequeue.append((other_label,))
        while dequeue:
            path = dequeue.popleft()
            end = path[-1]
            if self.equivalent(end, label):
                yield path
                continue
            for new_end in self.vertices[end]:
                if new_end in path:
                    continue
                dequeue.append(path + (new_end,))

    def _set_equivalent(self, label: int, other_label: int) -> None:
        """Set the two labels as equivalent in UnionFind"""
        verified = self.is_verified(label) or self.is_verified(other_label)
        roots = [self[label], self[other_label]]
        heaviest = max([(self.weights[r], r) for r in roots])[1]
        for r in roots:
            if r != heaviest:
                self.weights[heaviest] += self.weights[r]
                self.parents[r] = heaviest
        if verified:
            self.set_verified(label)

    def equivalent(self, label: int, other_label: int) -> bool:
        """Return True if label and other_label are equivalent, False otherwise."""
        return self[label] == self[other_label]

    def set_verified(self, comb_class: int) -> None:
        """Update database that combinatorial classes equivalent to comb_class
        are verified."""
        if not self.is_verified(comb_class):
            self.verified_roots.add(self[comb_class])

    def is_verified(self, comb_class: int) -> bool:
        """Return true if any equivalent combinatorial class is verified."""
        return self[comb_class] in self.verified_roots

    def equivalent_set(self, comb_class: int) -> Set[int]:
        """Return all of the classes equivalent to comb_class."""
        equivalent_classes = set()
        for t in self.parents:
            if self.equivalent(comb_class, t):
                equivalent_classes.add(t)
        return equivalent_classes

    def find_path(self, comb_class: int, other_comb_class: int) -> Tuple[int, ...]:
        """
        BFS for shortest path from comb_class to other_comb_class.

        Used to find shortest explanation of why things are equivalent.
        """
        if not self.equivalent(comb_class, other_comb_class):
            raise KeyError("The classes given are not equivalent.")

        dequeue: Deque[Tuple[int, ...]] = deque()
        dequeue.append((comb_class,))
        visited: Set[int] = set()
        while dequeue:
            path = dequeue.popleft()
            end = path[-1]
            if end == other_comb_class:
                break
            if end in visited:
                continue
            for new_end in self.vertices[end]:
                if new_end in path:
                    continue
                dequeue.append(path + (new_end,))
            visited.add(path[-1])
        return path



import itertools
from collections import deque
from operator import itemgetter
from typing import TYPE_CHECKING, Deque, Dict, Iterator, List, Set, Tuple

from comb_spec_searcher.class_db import ClassDB
from comb_spec_searcher.strategies.rule import AbstractRule, Rule
from comb_spec_searcher.tree_searcher import Node
from comb_spec_searcher.typing import RuleKey

if TYPE_CHECKING:
    from comb_spec_searcher.rule_db.base import RuleDBBase


class SpecificationRuleExtractor:
    def __init__(
        self, root_label: int, root_node: Node, ruledb: "RuleDBBase", classdb: ClassDB
    ):
        self.ruledb = ruledb
        self.classdb = classdb
        self.root_label = root_label
        self.eqv_rulekeys = root_node.rule_keys()
        self.rules_dict: Dict[int, Tuple[int, ...]] = {}
        # A map from equiv label to an equivalent label actually in the tree.
        self.eqvparent_to_parent: Dict[int, int] = {}
        self._populate_decompositions()
        self._populate_equivalences()
        self._check()

    def _populate_decompositions(self) -> None:
        """
        Populate the rules dict with the labels of decomposition rules.
        """
        eqvrule_to_rule = self.ruledb.rule_from_equivalence_rule_dict(self.eqv_rulekeys)
        for eqvrule in self.eqv_rulekeys:
            parent, children = eqvrule_to_rule[eqvrule]
            self.rules_dict[parent] = children
            self.eqvparent_to_parent[eqvrule[0]] = parent

    def _populate_equivalences(self) -> None:
        """
        Populate the rules dict with the labels of equivalences rules needed to make
        every class be on a right hand side.
        """
        for label in self._no_lhs_labels():
            eqv_label = self.ruledb.equivdb[label]
            path = self.ruledb.equivdb.find_path(
                label, self.eqvparent_to_parent[eqv_label]
            )
            for parent, child in zip(path[:-1], path[1:]):
                if parent in self.rules_dict:
                    # No need to keep adding the path since the class already connects
                    # with something in the specification
                    break
                self.rules_dict[parent] = (child,)

    def _no_lhs_labels(self) -> Set[int]:
        """
        Find labels that are not a lhs of any rule.
        """
        all_rhs = itertools.chain.from_iterable(self.rules_dict.values())
        no_lhs_labels = set(
            itertools.filterfalse(self.rules_dict.__contains__, all_rhs)
        )
        if self.root_label not in self.rules_dict:
            no_lhs_labels.add(self.root_label)
        return no_lhs_labels

    def _check(self):
        """
        Perform a check to make sure that the spec make sense.
        """
        all_rhs = set(
            itertools.chain.from_iterable(
                children for children in self.rules_dict.values()
            )
        )
        assert all_rhs.issubset(self.rules_dict), "Something is not on the lhs"
        extra_lhs = {c for c in self.rules_dict if c not in all_rhs}
        assert extra_lhs.issubset({self.root_label})

    def _find_rule(self, parent: int, children: Tuple[int, ...]) -> AbstractRule:
        try:
            strategy = self.ruledb.rule_to_strategy[(parent, children)]
        except KeyError as e:
            if len(children) != 1:
                raise self._missing_rule_error(parent, children) from e
        else:
            return strategy(self.classdb.get_class(parent))
        # From now on we are looking for a two way strategy
        assert len(children) == 1
        try:
            strategy = self.ruledb.eqv_rule_to_strategy[(parent, children)]
        except KeyError:
            pass
        else:
            rule = strategy(self.classdb.get_class(parent))
            assert isinstance(rule, Rule)
            return rule if len(rule.children) == 1 else rule.to_equivalence_rule()
        # Now trying a reverse rule
        try:
            strategy = self.ruledb.eqv_rule_to_strategy[(children[0], (parent,))]
        except KeyError:
            pass
        else:
            rule = strategy(self.classdb.get_class(children[0]))
            assert isinstance(rule, Rule)
            rule = rule if len(rule.children) == 1 else rule.to_equivalence_rule()
            return rule.to_reverse_rule(0)
        raise self._missing_rule_error(parent, children)

    def _missing_rule_error(self, parent: int, children: Tuple[int, ...]) -> ValueError:
        """
        Return a ValueError saying the rule was not found.
        """
        msg = f"Unable to retrieve rule for {(parent, children)}\n"
        msg += f"Class {parent} is:\n{self.classdb.get_class(parent)}\n"
        for child in children:
            msg += f"Class {child} is:\n{self.classdb.get_class(child)}\n"
        return ValueError(msg)

    def rules(self) -> Iterator[AbstractRule]:
        for parent, children in self.rules_dict.items():
            yield self._find_rule(parent, children)


class PartialSpecificationRuleExtractor(SpecificationRuleExtractor):
    """A base class for extractors for incomplete specifications."""

    def _populate_decompositions(self) -> None:
        pass

    def _populate_equivalences(self) -> None:
        # SpecificationRuleExtractor's with a try-catch do deal with incomplete spec
        for label in self._no_lhs_labels():
            eqv_label = self.ruledb.equivdb[label]
            # Might not exist because spec is incomplete.
            try:
                path = self.ruledb.equivdb.find_path(
                    label, self.eqvparent_to_parent[eqv_label]
                )
            except KeyError:
                continue
            for parent, child in zip(path[:-1], path[1:]):
                if parent in self.rules_dict:
                    break
                self.rules_dict[parent] = (child,)

    def _check(self):
        pass

    def _ordered_eqvrule_to_rule(
        self,
    ) -> Tuple[Dict[RuleKey, RuleKey], Dict[RuleKey, Tuple[int, ...]]]:
        """Altered version of rule_from_equivalence_rule_dict that deals with order."""
        eqv_rules = set(self.eqv_rulekeys)
        eqvrule_to_rule: Dict[RuleKey, RuleKey] = {}
        # Maps our children index order to that of the actual rules
        index_order_map: Dict[RuleKey, Tuple[int, ...]] = {}
        for start, ends in self.ruledb.rule_to_strategy:
            eqv_start = self.ruledb.equivdb[start]
            # Sort but keep knowledge of original order
            index_order, eqv_ends = (
                zip(
                    *sorted(
                        enumerate(map(self.ruledb.equivdb.__getitem__, ends)),
                        key=itemgetter(1),
                    )
                )
                if ends
                else ((), ())
            )

            eqv_key = (eqv_start, eqv_ends)
            index_order_map[eqv_key] = index_order
            if eqv_key in eqv_rules:
                eqvrule_to_rule[eqv_key] = (start, ends)
        return eqvrule_to_rule, index_order_map


class EquivalenceRuleExtractor(PartialSpecificationRuleExtractor):
    """
    This class extracts the actual rules that would be created for an equivalence
    label given a partial spec that is traversable from root. Suppose we want the
    rules for the parent in `grandparent-parent-children` than what rules are used
    will depend on the grandparent and children and we are looking for the rules
    that have the same equivalance label on both the LHS and RHS and in the order they
    would be traversed from the root.
    """

    def __init__(
        self,
        root_eq_label: int,
        root_class_label: int,
        root_node: Node,
        ruledb: "RuleDBBase",
        classdb: ClassDB,
        target: int,
        parent_of_target: int,
        idx: int,
    ):
        # parent
        self.target = target
        # grandparent
        self.parent_of_target = parent_of_target
        # grandparent.children[idx] is parent
        self.idx = idx
        # class label of the LHS of the first rule for the eq path, updated later
        self.start = root_class_label
        # class label of the RHS of the last rule for the eq path, updated later
        self.end = -1
        # Children are included in the tree
        super().__init__(root_eq_label, root_node, ruledb, classdb)

    def _populate_decompositions(self) -> None:
        eqvrule_to_rule, index_order_map = self._ordered_eqvrule_to_rule()
        for eqvrule in self.eqv_rulekeys:
            # Might not exist because spec is incomplete.
            try:
                parent, children = eqvrule_to_rule[eqvrule]
            except KeyError:
                continue
            self.rules_dict[parent] = children
            self.eqvparent_to_parent[eqvrule[0]] = parent
            # If grandparent, then we can find our starting point
            # If no grandparent exists, this will always be the root.
            if eqvrule[0] == self.parent_of_target:
                self.start = children[index_order_map[eqvrule][self.idx]]
            # If parent, we have our end point
            if eqvrule[0] == self.target:
                self.end = parent

    def nonequivalent_rules_in_equiv_path(self) -> List[Rule]:
        return list(self._nonequivalent_rules_in_equiv_path())

    def _nonequivalent_rules_in_equiv_path(self) -> Iterator[Rule]:
        path = self.ruledb.equivdb.find_path(self.start, self.end)
        for p, c in zip(path, path[1:]):
            rule = self._find_rule(p, (c,))
            # If rule is an actual equivalence rule, we don't care and ignore it.
            if not rule.is_equivalence():
                assert isinstance(rule, Rule)
                yield rule


class RulePathToAtomExtractor(PartialSpecificationRuleExtractor):
    """
    This class extracts rules from the root down to an atom in a specification (in terms
    of eq label maps) that is not complete but contains a path from root to said atom.
    """

    def __init__(
        self,
        root_class_label: int,
        root_label: int,
        root_node: Node,
        ruledb: "RuleDBBase",
        classdb: ClassDB,
        path: List[Tuple[int, int]],
    ):
        self.root_class_label = root_class_label
        self.path = path
        self.path.pop()  # remove padded -1
        self.search_order: Dict[int, Tuple[int, ...]] = {}
        super().__init__(root_label, root_node, ruledb, classdb)

    def _populate_decompositions(self) -> None:
        eqvrule_to_rule, index_order_map = self._ordered_eqvrule_to_rule()
        for eqvrule in self.eqv_rulekeys:
            try:
                parent, children = eqvrule_to_rule[eqvrule]
            except KeyError:
                continue
            self.rules_dict[parent] = children
            self.eqvparent_to_parent[eqvrule[0]] = parent

            # Order in searcher
            order = index_order_map[eqvrule]
            self.search_order[parent] = tuple(
                children[order[i]] for i in range(len(children))
            )

    def rule_path(self) -> Deque[Tuple[AbstractRule, int]]:
        curr_class = self.root_class_label
        path: Deque[Tuple[AbstractRule, int]] = deque([])
        while self.path:
            children = self.rules_dict[curr_class]
            if (
                len(children) == 1
                and self.ruledb.equivdb[curr_class] == self.ruledb.equivdb[children[0]]
            ):
                # If rule is within eq class, we add it with idx 0
                path.append((self._find_rule(curr_class, children), 0))
                curr_class = children[0]
            else:
                _, idx = self.path.pop()
                rule = self._find_rule(curr_class, children)
                # The idx given by the searcher might differ from that of rule_dict
                # and we use search_order to get the corresponding index in rule_dict
                next_child = self.search_order[curr_class][idx]
                # Get the first child that matches the label of the child we want.
                # If there are multiple, it does not matter which one.
                rule_idx = next(
                    i
                    for i, c in enumerate(rule.children)
                    if self.classdb.get_label(c) == next_child
                )
                path.append((rule, rule_idx))
                curr_class = next_child
        return path



"""
A combinatorial specification is a set rules of the form a -> b1, ..., bk
where each of the bi appear exactly once on the left hand side of some rule.
"""
from copy import copy
from functools import reduce
from operator import mul
from typing import Dict, Generic, Iterable, Iterator, List, Optional, Set, Union

import sympy
from logzero import logger
from sympy import Eq, Expr, Function, Number, solve, var

from comb_spec_searcher.class_queue import DefaultQueue
from comb_spec_searcher.exception import SpecificationNotFound
from comb_spec_searcher.typing import (
    CombinatorialClassType,
    CombinatorialObjectType,
    Objects,
    Terms,
)

from .combinatorial_class import CombinatorialClass, CombinatorialObject
from .exception import (
    IncorrectGeneratingFunctionError,
    InvalidOperationError,
    TaylorExpansionError,
)
from .isomorphism import Bijection, Isomorphism
from .specification_drawer import SpecificationDrawer
from .strategies import (
    EmptyStrategy,
    EquivalencePathRule,
    Rule,
    StrategyPack,
    VerificationRule,
    VerificationStrategy,
)
from .strategies.rule import AbstractRule
from .utils import (
    RecursionLimit,
    maple_equations,
    pretty_print_equations,
    taylor_expand,
)

__all__ = ("CombinatorialSpecification",)


class CombinatorialSpecification(
    Generic[CombinatorialClassType, CombinatorialObjectType]
):
    """
    A combinatorial specification is a set rules of the form a -> b1, ..., bk
    where each of the bi appear exactly once on the left hand side of some
    rule.
    """

    def __init__(
        self,
        root: CombinatorialClassType,
        rules: Iterable[AbstractRule[CombinatorialClassType, CombinatorialObjectType]],
        group_equiv: bool = True,
    ):
        self.root = root
        self.rules_dict = {rule.comb_class: rule for rule in rules}
        self._class_to_label: Dict[CombinatorialClassType, int] = {}
        self._label_to_class: Dict[int, CombinatorialClassType] = {}
        if group_equiv:
            self._group_equiv_in_path()
        self._set_subrules()
        self._enforce_labels()

    def _enforce_labels(self) -> None:
        """
        Traverses the spec in a DFS order from the root and calls self.get_label on
        each to assign labels to each node.
        """
        done = set()
        todo = [self.root]

        while todo:
            class_to_process = todo.pop()
            self.get_label(class_to_process)
            done.add(class_to_process)
            children = self.rules_dict[class_to_process].children
            todo.extend([child for child in children[::-1] if child not in done])

    def _set_subrules(self) -> None:
        """Tells the subrules which children's recurrence methods it should use."""
        for rule in list(self):  # list as we lazily assign empty rules
            rule.set_subrecs(self.get_rule)

    def _group_equiv_in_path(self) -> None:
        """
        Group chain of equivalence rules of the specification into equivalence paths.
        """
        self._ungroup_equiv_path()
        not_hidden_classes = {self.root}
        for rule in self:
            if rule.is_equivalence():
                continue
            not_hidden_classes.add(rule.comb_class)
            not_hidden_classes.update(rule.children)
        # Creating paths
        comb_class_stack = [self.root]
        visited: Set[CombinatorialClassType] = set()
        path_rules: List[Rule] = []
        eqv_path_rules: Dict[CombinatorialClassType, EquivalencePathRule] = {}
        while comb_class_stack:
            if path_rules and path_rules[-1].children[0] in not_hidden_classes:
                new_rule: EquivalencePathRule = EquivalencePathRule(path_rules)
                eqv_path_rules[new_rule.comb_class] = new_rule
                path_rules.clear()
            comb_class = comb_class_stack.pop()
            if comb_class in not_hidden_classes and comb_class in visited:
                continue
            visited.add(comb_class)
            rule = self.get_rule(comb_class)
            comb_class_stack.extend(rule.children)
            if rule.is_equivalence() and not isinstance(rule, EquivalencePathRule):
                assert isinstance(rule, Rule)
                assert not path_rules or path_rules[-1].children[0] == rule.comb_class
                path_rules.append(rule)
        # Clearing the equivalences and replacing by paths
        self.rules_dict = {
            comb_class: rule
            for comb_class, rule in self.rules_dict.items()
            if comb_class in not_hidden_classes
        }
        self.rules_dict.update(eqv_path_rules)
        assert self._is_valid_spec()

    def _ungroup_equiv_path(self) -> None:
        """
        Remove the equiv path and replacing only with normal rules.
        """
        new_rules: Dict[CombinatorialClassType, Rule] = {}
        for rule in self:
            if not isinstance(rule, EquivalencePathRule):
                continue
            for r in rule.rules:
                new_rules[r.comb_class] = r
        self.rules_dict.update(new_rules)

    def expand_verified(
        self,
    ) -> "CombinatorialSpecification[CombinatorialClassType, CombinatorialObjectType]":
        """
        Will expand all verified classes with respect to the strategy packs
        given by the VerificationStrategies.
        """
        new_spec = self
        while True:
            try:
                class_to_expand = next(new_spec.unexpanded_verified_classes())
            except StopIteration:
                return new_spec
            rule = new_spec.rules_dict[class_to_expand]
            assert isinstance(rule, VerificationRule)
            pack = rule.pack()
            try:
                logger.info("Expanding with %s on \n%s\n", pack.name, class_to_expand)
                new_spec = new_spec.expand_comb_class(
                    class_to_expand,
                    pack,
                    reverse=False,
                    continue_expanding_verified=False,
                )
            except SpecificationNotFound:
                logger.info("Specification NOT detected. Allowing reverse rules")
                new_spec = new_spec.expand_comb_class(
                    class_to_expand,
                    pack,
                    reverse=True,
                    continue_expanding_verified=True,
                )

    def unexpanded_verified_classes(self) -> Iterator[CombinatorialClassType]:
        """
        Returns the verified classes so you can determine which classes still
        need to be expanded.
        """
        for comb_class, rule in self.rules_dict.items():
            if isinstance(rule, VerificationRule):
                try:
                    rule.pack()
                except InvalidOperationError:
                    continue
                yield comb_class

    def expand_comb_class(
        self,
        comb_class: Union[int, CombinatorialClassType],
        pack: StrategyPack,
        reverse: bool,
        continue_expanding_verified: bool,
        max_expansion_time: Optional[float] = None,
    ) -> "CombinatorialSpecification[CombinatorialClassType, CombinatorialObjectType]":
        """
        Will try to expand a particular class with respect to the given strategy pack.

        If reverse is set to True, will allow to use reverse rule in the expansion.

        If continue continue_expanding_verified is set to True then the css will keep
        working on verified classes when performing the search.
        """
        # pylint: disable=import-outside-toplevel
        from .comb_spec_searcher import CombinatorialSpecificationSearcher
        from .rule_db import RuleDBForest

        if isinstance(comb_class, int):
            comb_class = self.get_comb_class(comb_class)

        spec_rules = tuple(
            copy(rule) for cc, rule in self.rules_dict.items() if cc != comb_class
        )
        ruledb = RuleDBForest(reverse=False, rule_cache=spec_rules)
        css = CombinatorialSpecificationSearcher(
            self.root,
            pack,
            ruledb=ruledb,
            debug=False,
            expand_verified=continue_expanding_verified,
        )
        for rule in spec_rules:
            start_label = css.classdb.get_label(rule.comb_class)
            end_labels = tuple(map(css.classdb.get_label, rule.children))
            ruledb.add(start_label, end_labels, rule)
        ruledb.reverse = reverse
        css.classqueue = DefaultQueue(css.strategy_pack)
        label_to_expand = css.classdb.get_label(comb_class)
        css.classqueue.add(label_to_expand)
        css.try_verify(comb_class, label_to_expand)
        # logger.info(CSS.run_information())
        try:
            # pylint: disable=protected-access
            spec_rule = css._auto_search_rules(max_expansion_time=max_expansion_time)
        except SpecificationNotFound as e:
            raise SpecificationNotFound("Expansion unsuccessful") from e
        new_spec = CombinatorialSpecification(self.root, spec_rule)
        return new_spec

    def _is_valid_spec(self) -> bool:
        """Checks that each class is on a left hand side."""
        comb_classes = set()
        for rule in self:
            comb_classes.add(rule.comb_class)
            comb_classes.update(rule.children)
        return self.root in comb_classes and all(
            c in self.rules_dict or c.is_empty() for c in comb_classes
        )

    def get_rule(
        self, comb_class: Union[int, CombinatorialClassType]
    ) -> AbstractRule[CombinatorialClassType, CombinatorialObjectType]:
        """Return the rule with comb class on the left."""
        if isinstance(comb_class, int):
            comb_class = self.get_comb_class(comb_class)
        if comb_class not in self.rules_dict:
            assert (
                comb_class.is_empty()
            ), f"rule not in the spec and not empty\n{comb_class}"
            empty_strat = EmptyStrategy[
                CombinatorialClassType, CombinatorialObjectType
            ]()
            self.rules_dict[comb_class] = empty_strat(comb_class)
        return self.rules_dict[comb_class]

    def comb_classes(self) -> Set[CombinatorialClassType]:
        """
        Return a set containing all the combinatorial classes of the specification.
        """
        res: Set[CombinatorialClassType] = set()
        for rule in self:
            res.update(rule.children)
            res.add(rule.comb_class)
            if isinstance(rule, EquivalencePathRule):
                for parent, _ in rule.eqv_path_rules():
                    res.add(parent)
        return res

    @property
    def root_rule(
        self,
    ) -> AbstractRule[CombinatorialClassType, CombinatorialObjectType]:
        """Return the rule of the root comb class."""
        return self.rules_dict[self.root]

    def get_label(self, comb_class: CombinatorialClassType) -> int:
        """Return a unique label for the comb class."""
        res = self._class_to_label.get(comb_class)
        if res is None:
            res = len(self._class_to_label)
            self._class_to_label[comb_class] = res
            self._label_to_class[res] = comb_class
        return res

    def get_comb_class(self, label: int) -> CombinatorialClassType:
        """Return the comb class associated to the unique label."""
        try:
            comb_class = self._label_to_class[label]
        except KeyError as e:
            raise InvalidOperationError(
                f"The label {comb_class} does not correspond to a tiling"
                " in the specification."
            ) from e
        return comb_class

    def get_function(self, comb_class: CombinatorialClassType) -> Function:
        """
        Return a sympy function for the comb class, using the label it is
        assigned.
        """
        return comb_class.get_function(self.get_label)

    def get_equations(self) -> Iterator[Eq]:
        """
        Yield all equations on the (ordinary) generating function that the
        rules of the specification imply.
        """
        funcs: Dict[CombinatorialClass, Function] = {
            comb_class: self.get_function(comb_class)
            for comb_class, rule in self.rules_dict.items()
            if not isinstance(rule, VerificationRule)
        }
        for _, rule in sorted(
            self.rules_dict.items(), key=lambda x: self.get_label(x[0])
        ):
            try:
                if isinstance(rule, VerificationRule):
                    eq = rule.get_equation(self.get_function, funcs)
                else:
                    eq = rule.get_equation(self.get_function)
                if not isinstance(eq, bool):
                    yield eq
            except NotImplementedError:
                logger.info(
                    "can't find generating function for the rule %s -> %s. "
                    "The rule was:\n%s",
                    self.get_label(rule.comb_class),
                    tuple(self.get_label(child) for child in rule.children),
                    rule,
                )
                x = var("x")
                yield Eq(
                    self.get_function(rule.comb_class),
                    sympy.Function("NOTIMPLEMENTED")(x),
                )

    def get_initial_conditions(self, check: int = 6) -> List[Expr]:
        """
        Compute the initial conditions of the root class. It will use the
        `count_objects_of_size` method if implemented, else resort to
        the method on the `initial_conditions` method on `CombinatorialClass.
        """
        logger.info("Computing initial conditions")
        try:
            return [
                sum(
                    Number(self.count_objects_of_size(n=n, **parameters))
                    * reduce(mul, [var(k) ** val for k, val in parameters.items()], 1)
                    for parameters in self.root.possible_parameters(n)
                )
                for n in range(check + 1)
            ]
        except NotImplementedError as e:
            logger.info(
                "Reverting to generating objects from root for initial "
                "conditions due to:\nNotImplementedError: %s",
                e,
            )
        return self.root.initial_conditions(check)

    def get_genf(self, check: int = 6) -> Expr:
        """
        Return the generating function for the root comb class.

        # TODO: consider what to do if multiple variables.
        """
        eqs = tuple(self.get_equations())
        root_func = self.get_function(self.root)
        logger.info("Computing initial conditions")
        initial_conditions = self.get_initial_conditions(check)
        logger.info(pretty_print_equations(root_func, initial_conditions, eqs))
        logger.info("Solving...")
        solutions = solve(
            eqs,
            tuple(eq.lhs for eq in eqs),
            dict=True,
            cubics=False,
            quartics=False,
            quintics=False,
        )
        for solution in solutions:
            genf = solution[root_func]
            logger.info("Checking initial conditions for: %s", genf)
            try:
                expansion = taylor_expand(genf, check)
            except TaylorExpansionError:
                continue
            if expansion == initial_conditions:
                return sympy.simplify(genf)
        raise IncorrectGeneratingFunctionError(
            "Failed to compute the generating function for the specification."
        )

    def get_maple_equations(self, check: int = 6) -> str:
        """
        Convert the systems of equations to version that can be copy pasted to maple.
        """
        eqs = tuple(self.get_equations())
        root_func = self.get_function(self.root)
        initial_conditions = self.get_initial_conditions(check)
        maple_eqs = maple_equations(root_func, initial_conditions, eqs)
        return maple_eqs

    def get_pretty_equations(self, check: int = 6) -> str:
        """
        Convert the systems of equations to a more readable format.
        """
        eqs = tuple(self.get_equations())
        root_func = self.get_function(self.root)
        initial_conditions = self.get_initial_conditions(check)
        return pretty_print_equations(root_func, initial_conditions, eqs)

    def count_objects_of_size(self, n: int, **parameters) -> int:
        """
        Return the number of objects with the given parameters.
        Note, 'n' is reserved for the size of the object.
        """
        limit = n * self.number_of_rules()
        with RecursionLimit(limit):
            return self.root_rule.count_objects_of_size(n, **parameters)

    def get_terms(self, n: int) -> Terms:
        """
        Return the terms for given n.
        """
        return self.root_rule.get_terms(n)

    def get_objects(self, n: int) -> Objects:
        """
        Return the objects for given n.
        """
        return self.root_rule.get_objects(n)

    def generate_objects_of_size(
        self, n: int, **parameters
    ) -> Iterator[CombinatorialObjectType]:
        """
        Return the objects with the given parameters.
        Note, 'n' is reserved for the size of the object.
        """
        for obj in self.root_rule.generate_objects_of_size(n, **parameters):
            yield obj

    def random_sample_object_of_size(
        self, n: int, **parameters: int
    ) -> CombinatorialObjectType:
        """
        Return a uniformly random object of the given size. This is done using
        the "recursive" method.
        """
        limit = n * self.number_of_rules()
        with RecursionLimit(limit):
            if self.count_objects_of_size(n, **parameters) > 0:
                return self.root_rule.random_sample_object_of_size(n, **parameters)
            raise InvalidOperationError(
                "The root does not contain objects of this size"
            )

    def number_of_rules(self) -> int:
        return len(self.rules_dict)

    def sanity_check(self, length: int = 5) -> bool:
        """
        Sanity check the specification to the given length.

        Raise an SanityCheckFailure error if it fails.
        """
        if not self._is_valid_spec():
            return False

        for rule in self:
            try:
                for n in range(length + 1):
                    if not rule.sanity_check(n):
                        return False
            except NotImplementedError:
                logger.warning(
                    "Can't sanity check the rule %s -> %s, which is\n" "%s",
                    self.get_label(rule.comb_class),
                    tuple(self.get_label(child) for child in rule.children),
                    rule,
                )

        return True

    def get_bijection_to(
        self, other: "CombinatorialSpecification"
    ) -> Optional[Bijection]:
        """Get bijection from self to other."""
        return Bijection.construct(self, other)

    def are_isomorphic(self, other: "CombinatorialSpecification") -> bool:
        """Check if self is isomorphic to other."""
        return Isomorphism.check(self, other)

    def show(
        self, levels_shown: int = 0, levels_expand: int = 0, verbose: bool = False
    ) -> None:
        """
        Displays a tree representing this object in the web browser
        OTHER INPUT:
            - 'levels_shown': number of levels displayed at the start.
            If 0 then the whole tree is displayed
            - 'levels_expand': number of levels displayed after expanding a node.
            If 0 then the rest of the tree is displayed
            - 'verbose': displays formal step below the node
            and json representation in tooltips
        """
        sd = SpecificationDrawer(
            self,
            levels_shown=levels_shown,
            levels_expand=levels_expand,
            verbose=verbose,
        )
        sd.show()

    def share(
        self, levels_shown: int = 0, levels_expand: int = 0, verbose: bool = False
    ) -> None:
        """
        Uploads the html representation of the specification to a sharing website and
        display the link to the file.

        INPUT:
            - 'levels_shown': number of levels displayed at the start.
            If 0 then the whole tree is displayed
            - 'levels_expand': number of levels displayed after expanding a node.
            If 0 then the rest of the tree is displayed
            - 'verbose': displays formal step below the node
            and json representation in tooltips
        """
        sd = SpecificationDrawer(
            self,
            levels_shown=levels_shown,
            levels_expand=levels_expand,
            verbose=verbose,
        )
        file_url = sd.share()
        print(f"The specification is available at {file_url}.")

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, CombinatorialSpecification):
            return NotImplemented
        return bool(self.root == other.root and self.rules_dict == other.rules_dict)

    def __str__(self) -> str:
        res = f"A combinatorial specification with {self.number_of_rules()} rules."
        rules_dict = copy(self.rules_dict)

        def update_res(comb_class: CombinatorialClassType, res: str):
            try:
                rule = rules_dict.pop(comb_class)
            except KeyError:
                assert comb_class in self.rules_dict or comb_class.is_empty()
                return res
            if isinstance(rule, EquivalencePathRule):
                child_label = self.get_label(rule.comb_class)
                child_eqv_label = self.get_label(rule.children[0])
                labels = f"{child_label} = {child_eqv_label}\n"
                res += f"\n{'-' * (len(labels) - 1)}\n"
                res += labels
                res += str(rule)
                try:
                    rule = rules_dict.pop(rule.children[0])
                except KeyError:
                    assert comb_class in self.rules_dict
                    return res
            start_label = self.get_label(rule.comb_class)
            end_labels = tuple(self.get_label(c) for c in rule.children)
            labels = f"{start_label} -> {end_labels}\n"
            res += f"\n{'-' * (len(labels) - 1)}\n"
            res += labels
            res += str(rule)
            for child in rule.children:
                res = update_res(child, res)
            return res

        res = update_res(self.root, res)
        assert not rules_dict
        return res + "\n"

    def __iter__(
        self,
    ) -> Iterator[AbstractRule[CombinatorialClassType, CombinatorialObjectType]]:
        yield from self.rules_dict.values()

    def equations_string(self) -> str:
        """
        Return a convenient to read string version of the equations returned by
        'get_equations' method.
        """
        res = ""
        for eq in sorted(
            self.get_equations(),
            key=lambda eq: int(str(eq.lhs).split("_")[1].split("(")[0]),
        ):
            res += f"\n{eq.lhs} = {eq.rhs}"
        return res

    # JSON methods

    def to_jsonable(self) -> dict:
        """Return a JSON serializable dictionary for the specification."""
        return {
            "root": self.root.to_jsonable(),
            "rules": [rule.to_jsonable() for rule in self],
        }

    @classmethod
    def from_dict(cls, d):
        """
        Return the specification with the dictionary outputter by the
        'to_jsonable' method
        """
        root = CombinatorialClass.from_dict(d.pop("root"))
        rules = [AbstractRule.from_dict(rule_dict) for rule_dict in d.pop("rules")]
        return CombinatorialSpecification(root, rules, group_equiv=False)


class AlreadyVerified(VerificationStrategy[CombinatorialClass, CombinatorialObject]):
    def __init__(self, verfied_comb_classes: Iterable[CombinatorialClass]):
        self.verified_classes = frozenset(verfied_comb_classes)
        super().__init__()

    def verified(self, comb_class: CombinatorialClass) -> bool:
        return comb_class in self.verified_classes

    @staticmethod
    def formal_step() -> str:
        return "already verified"

    def to_jsonable(self) -> dict:
        d: dict = super().to_jsonable()
        d["comb_classes"] = [c.to_jsonable() for c in self.verified_classes]
        return d

    @classmethod
    def from_dict(cls, d: dict) -> "AlreadyVerified":
        return cls([CombinatorialClass.from_dict(c) for c in d["comb_classes"]])



import enum
from typing import (
    TYPE_CHECKING,
    Callable,
    Counter,
    DefaultDict,
    Dict,
    List,
    NamedTuple,
    Set,
    Tuple,
    TypeVar,
    Union,
)

if TYPE_CHECKING:
    from comb_spec_searcher.combinatorial_class import CombinatorialClass  # noqa: F401
    from comb_spec_searcher.combinatorial_class import CombinatorialObject  # noqa: F401
    from comb_spec_searcher.strategies.strategy import AbstractStrategy, StrategyFactory

__all__ = [
    "CombinatorialObjectType",
    "CombinatorialClassType",
    "ClassKey",
    "Key",
    "CSSstrategy",
    "RulesDict",
    "WorkPacket",
    "RuleKey",
    "Parameters",
    "ParametersMap",
    "RelianceProfile",
    "RuleBucket",
    "ForestRuleKey",
    "Objects",
    "ObjectsCache",
    "Terms",
    "SubObjects",
    "SubRecs",
    "SubSamplers",
    "SubTerms",
]

CombinatorialObjectType = TypeVar(
    "CombinatorialObjectType", bound="CombinatorialObject"
)

CombinatorialClassType = TypeVar("CombinatorialClassType", bound="CombinatorialClass")


ClassKey = Union[bytes, CombinatorialClassType]
Key = Union[CombinatorialClassType, int]

CSSstrategy = Union["AbstractStrategy", "StrategyFactory"]
RulesDict = Dict[int, Set[Tuple[int, ...]]]


class WorkPacket(NamedTuple):
    label: int
    strategies: Tuple[CSSstrategy, ...]
    inferral: bool


@enum.unique
class RuleBucket(enum.Enum):
    UNDEFINED = enum.auto()
    VERIFICATION = enum.auto()
    EQUIV = enum.auto()
    NORMAL = enum.auto()
    REVERSE = enum.auto()


RuleKey = Tuple[int, Tuple[int, ...]]


class ForestRuleKey(NamedTuple):
    parent: int
    children: Tuple[int, ...]
    shifts: Tuple[int, ...]
    bucket: RuleBucket

    @property
    def key(self) -> RuleKey:
        return (self.parent, self.children)


# From constructor
Parameters = Tuple[int, ...]
ParametersMap = Callable[[Parameters], Parameters]
RelianceProfile = Tuple[Dict[str, Tuple[int, ...]], ...]
Objects = DefaultDict[Parameters, List[CombinatorialObjectType]]
ObjectsCache = List[Objects]
Terms = Counter[Parameters]  # all terms for a fixed n
SubObjects = Tuple[Callable[[int], Objects], ...]
SubRecs = Tuple[Callable[..., int], ...]
SubSamplers = Tuple[Callable[..., CombinatorialObjectType], ...]
SubTerms = Tuple[Callable[[int], Terms], ...]



"""
A database to search for tree.
"""
import abc
import itertools
from collections import defaultdict
from typing import Dict, Iterable, Iterator, List, MutableMapping, Optional, Set, Tuple

import sympy
import tabulate
from logzero import logger

from comb_spec_searcher.equiv_db import EquivalenceDB
from comb_spec_searcher.exception import InvalidOperationError
from comb_spec_searcher.rule_db.abstract import RuleDBAbstract, ensure_specification
from comb_spec_searcher.specification_extrator import SpecificationRuleExtractor
from comb_spec_searcher.strategies import AbstractStrategy, VerificationRule
from comb_spec_searcher.strategies.rule import AbstractRule
from comb_spec_searcher.tree_searcher import (
    Node,
    iterative_proof_tree_finder,
    iterative_prune,
    proof_tree_generator_dfs,
    prune,
    smallish_random_proof_tree,
)
from comb_spec_searcher.typing import RuleKey, RulesDict

__all__ = ["RuleDBBase", "RuleDB"]


class RuleDBBase(RuleDBAbstract):
    """
    A database to find specification using the pruning method.
    """

    def __init__(self) -> None:
        super().__init__()
        self.equivdb = EquivalenceDB()
        # Store the pruned dict. Should be set back to None every time the ruledb is
        # edited. Use the propertu for clean access.
        self._pruned_dict: Optional[RulesDict] = None

    @property
    def iterative(self) -> bool:
        """
        Indicate whether we are searching for an iterative specification.
        """
        return self.strategy_pack.iterative

    @property
    def pruned_dict(self) -> RulesDict:
        """
        Returned the prune dict of all the rules that are in a specification.

        The dict needs to be recomputed every time the database changes and it
        is costly to so. Call with moderation.
        """
        if self._pruned_dict is None:
            rules_dict = self.rules_up_to_equivalence()
            if self.iterative:
                rules_dict = iterative_prune(rules_dict, root=self.root_label)
            else:
                prune(rules_dict)
            self._pruned_dict = rules_dict
            for ver_label in rules_dict.keys():
                self.equivdb.set_verified(ver_label)
        return self._pruned_dict

    @property
    @abc.abstractmethod
    def rule_to_strategy(self) -> MutableMapping[RuleKey, AbstractStrategy]:
        pass

    @property
    @abc.abstractmethod
    def eqv_rule_to_strategy(self) -> MutableMapping[RuleKey, AbstractStrategy]:
        pass

    def __eq__(self, other: object) -> bool:
        """Check if all stored information is the same."""
        if not isinstance(other, self.__class__):
            return NotImplemented
        return (
            self.rule_to_strategy == other.rule_to_strategy
            and self.eqv_rule_to_strategy == other.eqv_rule_to_strategy
        )

    def add(self, start: int, ends: Tuple[int, ...], rule: AbstractRule) -> None:
        self._pruned_dict = None
        ends = self._clean_labels(ends, rule)
        if ends == [start]:
            return
        if isinstance(rule, VerificationRule):
            self.equivdb.set_verified(start)
        if len(ends) == 1:
            if rule.is_two_way():
                self.equivdb.add_two_way_edge(start, ends[0])
                self.eqv_rule_to_strategy[(start, ends)] = rule.strategy
                self.rule_to_strategy.pop((start, ends), None)
                self.rule_to_strategy.pop((ends[0], (start,)), None)
            else:
                self.equivdb.add_one_way_edge(start, ends[0])
                self.rule_to_strategy[(start, ends)] = rule.strategy
        else:
            self.rule_to_strategy[(start, ends)] = rule.strategy

    def _clean_labels(
        self, ends: Tuple[int, ...], rule: AbstractRule
    ) -> Tuple[int, ...]:
        """
        Remove the empty label and sort the remaining one.
        """
        cleaned_ends = []
        for comb_class, child_label in zip(rule.children, ends):
            if rule.possibly_empty and self.classdb.is_empty(comb_class, child_label):
                logger.debug("Label %s is empty.", child_label)
                self.searcher.classqueue.set_stop_yielding(child_label)
                continue
            cleaned_ends.append(child_label)
        return tuple(sorted(cleaned_ends))

    def is_verified(self, label: int) -> bool:
        """Return True if label has been verified."""
        return self.equivdb.is_verified(label)

    def are_equivalent(self, label: int, other: int) -> bool:
        """Return true if label and other are equivalent."""
        return self.equivdb.equivalent(label, other)

    def rules_up_to_equivalence(self) -> Dict[int, Set[Tuple[int, ...]]]:
        """Return a defaultdict containing all rules up to the equivalence."""
        self.equivdb.connect_cycles()
        rules_dict: Dict[int, Set[Tuple[int, ...]]] = defaultdict(set)
        for start, ends in self:
            if len(ends) == 1 and self.are_equivalent(start, ends[0]):
                continue
            rules_dict[self.equivdb[start]].add(
                tuple(sorted(self.equivdb[e] for e in ends))
            )
        return rules_dict

    def __iter__(self) -> Iterator[Tuple[int, Tuple[int, ...]]]:
        """Iterate through rules as the pairs (start, ends)."""
        return itertools.chain(self.rule_to_strategy, self.eqv_rule_to_strategy)

    def contains(self, start: int, ends: Tuple[int, ...]) -> bool:
        """Return true if the rule start -> ends is in the database."""
        key = (start, sorted(ends))
        return key in self.rule_to_strategy or key in self.eqv_rule_to_strategy

    def status(self, elaborate: bool) -> str:
        """Return a string describing the status of the rule database."""
        status = "RuleDB status:\n"
        table: List[Tuple[str, str]] = []
        table.append(("Combinatorial rules", f"{len(self.rule_to_strategy):,d}"))
        table.append(("Equivalence rules", f"{len(self.eqv_rule_to_strategy):,d}"))
        if elaborate:
            eqv_labels = set()
            strategy_verified_labels = set()
            verified_labels = set()
            eqv_rules = set()
            for start, ends in self.rule_to_strategy:
                if not ends:
                    strategy_verified_labels.add(start)
                if self.is_verified(start):
                    verified_labels.add(start)
                eqv_start = self.equivdb[start]
                eqv_ends = tuple(sorted(self.equivdb[label] for label in ends))
                eqv_labels.add(eqv_start)
                eqv_labels.update(eqv_ends)
                eqv_rules.add((eqv_start, eqv_ends))
            table.append(
                ("Combintorial rules up to equivalence", f"{len(eqv_rules):,d}")
            )
            table.append(
                (
                    "Strategy verified combinatorial classes",
                    f"{len(strategy_verified_labels):,d}",
                )
            )
            table.append(
                (
                    "Verified combinatorial classes",
                    f"{len(verified_labels):,d}",
                )
            )
            table.append(
                (
                    "combinatorial classes up to equivalence",
                    f"{len(eqv_labels):,d}",
                )
            )
        status += "    "
        status += tabulate.tabulate(
            table, headers=("", "Total number"), colalign=("left", "right")
        ).replace("\n", "\n    ")
        time_taken = round(self.equivdb.func_times["find_paths"], 2)
        status += (
            f"\n\tCalled find equiv path {self.equivdb.func_calls['find_paths']}"
            + f" times, for total time of {time_taken}"
            + " seconds.\n"
        )
        return status

    ################################################################
    # Below are methods for finding a combinatorial specification. #
    ################################################################

    def has_specification(self) -> bool:
        """Return True if a specification has been found, false otherwise."""
        return self.equivdb[self.root_label] in self.pruned_dict

    def rule_from_equivalence_rule(
        self, eqv_start: int, eqv_ends: Iterable[int]
    ) -> Optional[Tuple[int, Tuple[int, ...]]]:
        """
        Return a rule that satisfies the equivalence rule.

        Returns None if no such rule exists.
        """
        eqv_start = self.equivdb[eqv_start]
        eqv_ends = tuple(sorted(map(self.equivdb.__getitem__, eqv_ends)))
        for rule in self:
            start, ends = rule
            temp_start = self.equivdb[start]
            temp_ends = tuple(sorted(map(self.equivdb.__getitem__, ends)))
            if eqv_start == temp_start and eqv_ends == temp_ends:
                return start, ends
        return None

    def rule_from_equivalence_rule_dict(
        self, eqv_rules: Iterable[RuleKey]
    ) -> Dict[RuleKey, RuleKey]:
        """
        Return a dictionary pointing from an equivalence rule to an actual rule.
        """
        eqv_rules = set(eqv_rules)
        res: Dict[RuleKey, RuleKey] = {}
        for start, ends in self.rule_to_strategy:
            eqv_start = self.equivdb[start]
            eqv_ends = tuple(sorted(map(self.equivdb.__getitem__, ends)))
            eqv_key = (eqv_start, eqv_ends)
            if eqv_key in eqv_rules:
                res[eqv_key] = (start, ends)
        return res

    @ensure_specification
    def get_specification_rules(  # pylint: disable=arguments-differ
        self,
        *,
        minimization_time_limit: float = 10,
        smallest: bool = False,
        **kwargs,
    ) -> Iterator[AbstractRule]:
        node = self._get_specification_node(minimization_time_limit, smallest)
        logger.info("Found specification with %s rules.", len(node.labels()))
        spec_extractor = SpecificationRuleExtractor(
            self.root_label, node, self, self.classdb
        )
        return spec_extractor.rules()

    @ensure_specification
    def _get_specification_node(
        self, minimization_time_limit: float, smallest: bool
    ) -> Node:
        """
        Return a specification node for the given label.
        """
        if self.iterative:
            if smallest:
                raise InvalidOperationError("can't use iterative and smallest")
            node = self._get_iterative_node()
        else:
            if smallest:
                node = self._get_smallest_node(minimization_time_limit)
            else:
                node = self._get_smallish_node(
                    minimization_time_limit=minimization_time_limit,
                )
        return node

    @ensure_specification
    def _get_iterative_node(self) -> Node:
        if not self.iterative:
            raise InvalidOperationError("Not supported in non-iterative mode")
        return iterative_proof_tree_finder(
            self.pruned_dict, root=self.equivdb[self.root_label]
        )

    @ensure_specification
    def _get_smallish_node(self, minimization_time_limit: float) -> Node:
        """Search for a smallish node for the given minimization time."""
        if self.iterative:
            raise InvalidOperationError("Not supported in iterative mode.")
        logger.info("Minimizing for %s seconds.", round(minimization_time_limit))
        return smallish_random_proof_tree(
            self.pruned_dict, self.equivdb[self.root_label], minimization_time_limit
        )

    @ensure_specification
    def _get_smallest_node(self, minimization_time_limit: float) -> Node:
        """
        Return the smallest specification node in the universe.

        Will look for a smallish spec for minimization_time_limit and then use
        binary search to find it to find the smallest one.

        This doesn't consider the length of the equivalence paths.
        """
        if self.iterative:
            raise InvalidOperationError("Not supported in iterative mode.")
        node = self._get_smallish_node(minimization_time_limit)
        minimum = 1
        maximum = len(node)
        logger.info(
            "Found a specification of size %s. Looking for the smallest.", len(node)
        )
        # Binary search to find a smallest proof tree.
        while minimum < maximum:
            middle = (minimum + maximum) // 2
            logger.info("Looking for specification of size %s", middle)
            try:
                node = next(
                    proof_tree_generator_dfs(
                        self.pruned_dict,
                        root=self.equivdb[self.root_label],
                        maximum=middle,
                    )
                )
                maximum = min(middle, len(node))
            except StopIteration:
                minimum = middle + 1
        logger.info("The smallest specification is of size %s.", len(node))
        return node

    @ensure_specification
    def _all_nodes(self, iterative: bool = False) -> Iterator[Node]:
        """
        A generator that yields all specifications in the universe for
        the given label.
        """
        if self.iterative:
            raise NotImplementedError(
                "There is no method for yielding all iterative proof trees."
            )
        yield from proof_tree_generator_dfs(
            self.pruned_dict, root=self.equivdb[self.root_label]
        )


class RuleDB(RuleDBBase):
    def __init__(self) -> None:
        super().__init__()
        self._rule_to_strategy: Dict[Tuple[int, Tuple[int, ...]], AbstractStrategy] = {}
        self._eqv_rule_to_strategy: Dict[
            Tuple[int, Tuple[int, ...]], AbstractStrategy
        ] = {}

    @property
    def rule_to_strategy(self) -> Dict[Tuple[int, Tuple[int, ...]], AbstractStrategy]:
        return self._rule_to_strategy

    @property
    def eqv_rule_to_strategy(
        self,
    ) -> Dict[Tuple[int, Tuple[int, ...]], AbstractStrategy]:
        return self._eqv_rule_to_strategy

    def all_rules(self) -> Iterator[AbstractRule]:
        """Yield all the rules found so far."""
        for rulekey in self:
            parent = self.classdb.get_class(rulekey[0])
            try:
                strategy = self.rule_to_strategy[rulekey]
            except KeyError:
                strategy = self.eqv_rule_to_strategy[rulekey]
            yield strategy(parent)

    def all_equations(self) -> Set[sympy.Eq]:
        """
        Returns a set of equations for all rules currently found.
        """

        def get_function(comb_class) -> sympy.Function:
            return comb_class.get_function(self.classdb.get_label)

        eqs = set()
        for rule in self.all_rules():
            try:
                eq = rule.get_equation(get_function)
            except NotImplementedError:
                logger.info(
                    "can't find generating function for %s." " The comb class is:\n%s",
                    get_function(rule.comb_class),
                    rule.comb_class,
                )
                eq = sympy.Eq(
                    get_function(rule.comb_class),
                    sympy.Function("NOTIMPLEMENTED")(sympy.var("x")),
                )
            eqs.add(eq)
        return eqs



from collections import defaultdict
from random import randint
from typing import Callable, Counter, Dict, Iterator, List, Optional, Tuple

import sympy

from comb_spec_searcher.typing import (
    CombinatorialClassType,
    CombinatorialObjectType,
    Parameters,
    ParametersMap,
    RelianceProfile,
    SubObjects,
    SubRecs,
    SubSamplers,
    SubTerms,
    Terms,
)

from .base import Constructor


class DisjointUnion(Constructor[CombinatorialClassType, CombinatorialObjectType]):
    """
    The DisjointUnion constructor takes as input the children. Each constructor
    is unique up to the length of the children being used to count.

    Extra parameters are passed on using the parameters dictionaries. Each
    dictionary's keys should be the extra variable of the child pointing the
    variable on the parent it came from.
    If a parents variable does not map to a child, then this variable must be 0
    as the child contains no occurences.

    The fixed value dictionaries passed will be used ensure that the parameter
    of a child must take on the given value.
    """

    def __init__(
        self,
        parent: CombinatorialClassType,
        children: Tuple[CombinatorialClassType, ...],
        extra_parameters: Optional[Tuple[Dict[str, str], ...]] = None,
        fixed_values: Optional[Tuple[Dict[str, int], ...]] = None,
    ):
        self.number_of_children = len(children)
        if extra_parameters is not None:
            self.extra_parameters = extra_parameters
            assert len(extra_parameters) == len(children)
        else:
            assert not parent.extra_parameters
            self.extra_parameters = tuple({} for _ in range(self.number_of_children))
        self._children_param_maps = self._build_children_param_maps(parent, children)
        self.zeroes = tuple(
            frozenset(parent.extra_parameters) - frozenset(parameter.keys())
            for parameter in self.extra_parameters
        )
        if fixed_values is not None:
            self.fixed_values = fixed_values
            assert len(fixed_values) == len(children)
        else:
            self.fixed_values = tuple({} for _ in children)

    def get_equation(
        self, lhs_func: sympy.Function, rhs_funcs: Tuple[sympy.Function, ...]
    ) -> sympy.Eq:
        res = 0
        for rhs_func, extra_parameters in zip(rhs_funcs, self.extra_parameters):
            subs: Dict[str, sympy.Expr] = {}
            for parent, child in extra_parameters.items():
                if child in subs:
                    subs[child] *= sympy.var(parent)
                else:
                    subs[child] = sympy.var(parent)
            res += rhs_func.subs(
                subs,
                simultaneous=True,
            )
        return sympy.Eq(lhs_func, res)

    def reliance_profile(self, n: int, **parameters: int) -> RelianceProfile:
        assert not parameters, "only implemented in one variable, namely 'n'"
        return tuple({"n": (n,)} for _ in range(self.number_of_children))

    def get_extra_parameters(
        self, n: int, **parameters: int
    ) -> List[Optional[Dict[str, int]]]:
        """
        Will return the extra parameters dictionary based on the parent's
        parameters. If there is a contradiction, that is some child parameter
        is given two or more different values that do not match, then None
        will be returned for that child, indicating that 0 objects on the
        child match the parents parameters.
        """
        res: List[Optional[Dict[str, int]]] = []
        for i, extra_parameters in enumerate(self.extra_parameters):
            update_params: Dict[str, int] = {**self.fixed_values[i]}
            for parent_var, child_var in extra_parameters.items():
                updated_value = parameters[parent_var]
                if child_var not in update_params:
                    update_params[child_var] = updated_value
                elif update_params[child_var] != updated_value:
                    break
            else:
                res.append(update_params)
                continue
            res.append(None)
        return res

    def get_terms(
        self, parent_terms: Callable[[int], Terms], subterms: SubTerms, n: int
    ) -> Terms:
        new_terms: Terms = Counter()
        for child_terms, param_map in zip(subterms, self._children_param_maps):
            for param, value in child_terms(n).items():
                new_terms[param_map(param)] += value
        return new_terms

    def _build_children_param_maps(
        self,
        parent: CombinatorialClassType,
        children: Tuple[CombinatorialClassType, ...],
    ) -> Tuple[ParametersMap, ...]:
        map_list: List[ParametersMap] = []
        num_parent_params = len(parent.extra_parameters)
        parent_param_to_pos = {
            param: pos for pos, param in enumerate(parent.extra_parameters)
        }
        for child, extra_param in zip(children, self.extra_parameters):
            reversed_extra_param: Dict[str, List[str]] = defaultdict(list)
            for parent_var, child_var in extra_param.items():
                reversed_extra_param[child_var].append(parent_var)
            child_pos_to_parent_pos: Tuple[Tuple[int, ...], ...] = tuple(
                tuple(
                    map(
                        parent_param_to_pos.__getitem__,
                        reversed_extra_param[child_param],
                    )
                )
                if child_param in reversed_extra_param
                else tuple()
                for child_param in child.extra_parameters
            )
            map_list.append(
                self.build_param_map(child_pos_to_parent_pos, num_parent_params)
            )
        return tuple(map_list)

    def get_sub_objects(
        self, subobjs: SubObjects, n: int
    ) -> Iterator[
        Tuple[Parameters, Tuple[List[Optional[CombinatorialObjectType]], ...]]
    ]:
        res: List[List[Optional[CombinatorialObjectType]]]
        res = [[None] for _ in range(self.number_of_children)]
        for i, subobj in enumerate(subobjs):
            param_map = self._children_param_maps[i]
            for param, comb_objs in subobj(n).items():
                res[i] = comb_objs
                yield (param_map(param), tuple(res))
            res[i] = [None]

    def random_sample_sub_objects(
        self,
        parent_count: int,
        subsamplers: SubSamplers,
        subrecs: SubRecs,
        n: int,
        **parameters: int,
    ) -> Tuple[Optional[CombinatorialObjectType], ...]:
        random_choice = randint(1, parent_count)
        total = 0
        for (idx, rec), subsampler, extra_params in zip(
            enumerate(subrecs), subsamplers, self.get_extra_parameters(n, **parameters)
        ):
            # if a parent parameter is not mapped to by some child parameter
            # then it is assumed that the value of the parent parameter must be 0
            if extra_params is None or any(
                val != 0 and k in self.zeroes[idx] for k, val in parameters.items()
            ):
                continue
            total += rec(n=n, **extra_params)
            if random_choice <= total:
                obj = subsampler(n=n, **extra_params)
                return (
                    tuple(None for _ in range(idx))
                    + (obj,)
                    + tuple(None for _ in range(len(subrecs) - idx - 1))
                )

    @staticmethod
    def get_eq_symbol() -> str:
        return "="

    @staticmethod
    def get_op_symbol() -> str:
        return "+"

    def __str__(self):
        return "disjoint union"

    def equiv(
        self, other: "Constructor", data: Optional[object] = None
    ) -> Tuple[bool, Optional[object]]:
        return (
            isinstance(other, type(self))
            and DisjointUnion.extra_params_equiv(
                self.extra_parameters, other.extra_parameters
            ),
            None,
        )


class Complement(Constructor[CombinatorialClassType, CombinatorialObjectType]):
    """
    This constructor implements the complement operation. A rule A -> (B, C, D)
    should be read as A = B - C - D.

    This comes from a disjoint union rule B = A + C + D

    The initialiser should be made using the original disjoint union rule.

    The parent is B, idx is the index of the child the rule is being rearranged
    for and the extra parameters maps from parent to children.
    """

    def __init__(
        self,
        parent: CombinatorialClassType,
        children: Tuple[CombinatorialClassType, ...],
        idx: int,
        extra_parameters: Optional[Tuple[Dict[str, str], ...]] = None,
    ):
        self.number_of_children = len(children)
        if extra_parameters is not None:
            self.extra_parameters = extra_parameters
            assert len(extra_parameters) == len(children)
        else:
            assert not parent.extra_parameters
            self.extra_parameters = tuple({} for _ in range(self.number_of_children))
        self.idx = idx
        self._children_param_maps = self._build_children_param_maps(parent, children)
        self._parent_param_map = self._build_parent_param_map(parent, children, idx)

    def _build_children_param_maps(
        self,
        parent: CombinatorialClassType,
        children: Tuple[CombinatorialClassType, ...],
    ) -> Tuple[ParametersMap, ...]:
        map_list: List[ParametersMap] = []
        num_parent_params = len(parent.extra_parameters)
        parent_param_to_pos = {
            param: pos for pos, param in enumerate(parent.extra_parameters)
        }
        for (idx, child), extra_param in zip(
            enumerate(children), self.extra_parameters
        ):
            if idx == self.idx:
                continue
            reversed_extra_param: Dict[str, List[str]] = defaultdict(list)
            for parent_var, child_var in extra_param.items():
                reversed_extra_param[child_var].append(parent_var)
            child_pos_to_parent_pos: Tuple[Tuple[int, ...], ...] = tuple(
                tuple(
                    map(
                        parent_param_to_pos.__getitem__,
                        reversed_extra_param[child_param],
                    )
                )
                if child_param in reversed_extra_param
                else tuple()
                for child_param in child.extra_parameters
            )
            map_list.append(
                self.build_param_map(child_pos_to_parent_pos, num_parent_params)
            )
        return tuple(map_list)

    def _build_parent_param_map(
        self,
        parent: CombinatorialClassType,
        children: Tuple[CombinatorialClassType, ...],
        idx: int,
    ):
        extra_param = self.extra_parameters[idx]
        child = children[self.idx]
        child_param_to_pos = {
            param: pos for pos, param in enumerate(child.extra_parameters)
        }
        parent_pos_to_child_pos: Tuple[Tuple[int, ...], ...] = tuple(
            (child_param_to_pos[extra_param[parent_var]],)
            if parent_var in extra_param
            else tuple()
            for parent_var in parent.extra_parameters
        )
        return self.build_param_map(
            parent_pos_to_child_pos, len(child.extra_parameters)
        )

    def get_equation(
        self, lhs_func: sympy.Function, rhs_funcs: Tuple[sympy.Function, ...]
    ) -> sympy.Eq:
        if any(self.extra_parameters):
            raise NotImplementedError(
                "Complement equation is not implemented with extra parameters. "
                "You can fall back on the union equations."
            )
        res = rhs_funcs[0]
        for rhs_func in rhs_funcs[1:]:
            res -= rhs_func
        return sympy.Eq(lhs_func, res)

    def reliance_profile(self, n: int, **parameters: int) -> RelianceProfile:
        raise NotImplementedError

    def get_terms(
        self, parent_terms: Callable[[int], Terms], subterms: SubTerms, n: int
    ) -> Terms:
        parent_terms_mapped: Terms = Counter()
        for param, value in subterms[0](n).items():
            if value:
                parent_terms_mapped[self._parent_param_map(param)] += value
        children_terms = subterms[1:]
        for child_terms, param_map in zip(children_terms, self._children_param_maps):
            # we subtract from total
            for param, value in child_terms(n).items():
                mapped_param = self._parent_param_map(param_map(param))
                parent_terms_mapped[mapped_param] -= value
                assert parent_terms_mapped[mapped_param] >= 0
                if parent_terms_mapped[mapped_param] == 0:
                    parent_terms_mapped.pop(mapped_param)

        return parent_terms_mapped

    def get_sub_objects(
        self, subobjs: SubObjects, n: int
    ) -> Iterator[
        Tuple[Parameters, Tuple[List[Optional[CombinatorialObjectType]], ...]]
    ]:
        raise NotImplementedError

    def random_sample_sub_objects(
        self,
        parent_count: int,
        subsamplers: SubSamplers,
        subrecs: SubRecs,
        n: int,
        **parameters: int,
    ) -> Tuple[Optional[CombinatorialObjectType], ...]:
        raise NotImplementedError

    def __str__(self):
        return "complement"

    def equiv(
        self, other: "Constructor", data: Optional[object] = None
    ) -> Tuple[bool, Optional[object]]:
        return (
            isinstance(other, type(self))
            and Complement.extra_params_equiv(
                self.extra_parameters, other.extra_parameters
            ),
            None,
        )



"""A class for automatically performing combinatorial exploration."""
import gc
import logging
import platform
import time
import warnings
from collections import defaultdict
from datetime import timedelta
from typing import (
    Any,
    Dict,
    Generic,
    Iterator,
    List,
    Optional,
    Sequence,
    Set,
    Tuple,
    cast,
)

import logzero
import tabulate
from logzero import logger

from comb_spec_searcher.typing import CombinatorialClassType, CSSstrategy

from .class_db import ClassDB
from .class_queue import DefaultQueue
from .exception import (
    ExceededMaxtimeError,
    InvalidOperationError,
    SpecificationNotFound,
    StrategyDoesNotApply,
)
from .rule_db import RuleDB
from .rule_db.base import RuleDBAbstract
from .specification import CombinatorialSpecification
from .strategies import AbstractStrategy, StrategyFactory, StrategyPack
from .strategies.rule import AbstractRule
from .utils import (
    cssiteratortimer,
    cssmethodtimer,
    get_mem,
    nice_pypy_mem,
    size_to_readable,
)

if platform.python_implementation() == "CPython":
    from pympler.asizeof import asizeof

__all__ = ["CombinatorialSpecificationSearcher"]

warnings.simplefilter("once", Warning)

logzero.loglevel(logging.INFO)


class CombinatorialSpecificationSearcher(Generic[CombinatorialClassType]):
    """
    The CombinatorialSpecificationSearcher class.

    This is used to build up knowledge about a combinatorial_class with respect
    to the given strategies and search for a combinatorial specification.
    """

    def __init__(
        self,
        start_class: CombinatorialClassType,
        strategy_pack: StrategyPack,
        *,
        ruledb: Optional[RuleDBAbstract] = None,
        expand_verified: bool = False,
        debug: bool = False,
    ):
        """
        Initialise CombinatorialSpecificationSearcher.

        INPUTS:
          - `start_class`: the combinatorial class to search a specification for.
          - `strategy_pack`: a set of strategy to use for the search.
          - `ruledb`: a string to specify the type of ruledb to use for the
          search. Default to `None` but can be changed to "forget" for a ruledb that
          saves more memory.
          - `expand_verified`: if True, every verified combinatorial class will
            still be expanded using the strategies in strategy pack
          - `debug`: if True every rule found will be sanity checked and logged
            to logging.DEBUG
        """
        self.strategy_pack = strategy_pack
        self.debug = debug
        self.expand_verified = expand_verified
        if self.debug:
            logzero.loglevel(logging.DEBUG, True)

        self.func_times: Dict[str, float] = defaultdict(float)
        self.func_calls: Dict[str, int] = defaultdict(int)

        self.classdb = ClassDB[CombinatorialClassType](type(start_class))
        self.classqueue = DefaultQueue(strategy_pack)
        self.ruledb: RuleDBAbstract = ruledb if ruledb is not None else RuleDB()
        self.ruledb.link_searcher(self)

        # initialise the run with start_class
        self.start_label = self.classdb.get_label(start_class)
        self.classqueue.add(self.start_label)
        self.tried_to_verify: Set[int] = set()
        self.symmetry_expanded: Set[int] = set()
        self.inferral_expanded: Set[int] = set()
        self.try_verify(start_class, self.start_label)
        if self.symmetries:
            self._symmetry_expand(start_class, self.start_label)

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, CombinatorialSpecificationSearcher):
            return NotImplemented
        return self.__class__ == other.__class__ and self.__dict__ == other.__dict__

    @property
    def verification_strategies(self) -> Sequence[CSSstrategy]:
        """The verification strategies from the strategy pack."""
        return self.strategy_pack.ver_strats

    @property
    def symmetries(self) -> Sequence[CSSstrategy]:
        """The symmetries functions for the strategy pack."""
        return self.strategy_pack.symmetries

    @property
    def start_class(self) -> CombinatorialClassType:
        """Returns the start class of the searcher."""
        return self.classdb.get_class(self.start_label)

    def try_verify(self, comb_class: CombinatorialClassType, label: int) -> None:
        """
        Try to verify the combinatorial class.
        """
        if label in self.tried_to_verify:
            return
        self.tried_to_verify.add(label)
        if self.classdb.is_empty(comb_class, label):
            return
        for strategy in self.verification_strategies:
            if self.ruledb.is_verified(label):
                return
            for start_label, end_labels, rule in self._expand_class_with_strategy(
                comb_class, strategy, label
            ):
                self.add_rule(start_label, end_labels, rule)

    def _expand(
        self,
        comb_class: CombinatorialClassType,
        label: int,
        strategies: Tuple[CSSstrategy, ...],
        inferral: bool,
    ) -> None:
        """
        Will expand the combinatorial class with given label using the given
        strategies.
        """
        if inferral:
            self._inferral_expand(comb_class, label, strategies)
        else:
            for strategy_generator in strategies:
                for start_label, end_labels, rule in self._expand_class_with_strategy(
                    comb_class, strategy_generator, label
                ):
                    self.add_rule(start_label, end_labels, rule)

    @staticmethod
    def _rules_from_strategy(
        comb_class: CombinatorialClassType, strategy: CSSstrategy
    ) -> Iterator[AbstractRule]:
        """Yield all the rules given by a strategy/strategy factory."""
        if isinstance(strategy, AbstractStrategy):
            try:
                yield strategy(comb_class)
            except StrategyDoesNotApply:
                pass
        elif isinstance(strategy, StrategyFactory):
            for strat in strategy(comb_class):
                if isinstance(strat, AbstractRule):
                    yield strat
                elif isinstance(strat, AbstractStrategy):
                    try:
                        yield strat(comb_class)
                    except StrategyDoesNotApply:
                        continue
                else:
                    raise InvalidOperationError(
                        "Attempting to add non Rule type. A Strategy "
                        "Factory's __call__ method should yield strategy or "
                        "a rule."
                    )
        else:
            raise InvalidOperationError(
                "CSS can only expand a combinatorial class with "
                "Strategy and StrategyFactory"
            )

    @cssiteratortimer("_expand_class_with_strategy")
    def _expand_class_with_strategy(
        self,
        comb_class: CombinatorialClassType,
        strategy_generator: CSSstrategy,
        label: Optional[int] = None,
        initial: bool = False,
    ) -> Iterator[Tuple[int, Tuple[int, ...], AbstractRule]]:
        """
        Will expand the class with given strategy. Return time taken.
        """
        logger.debug(
            "Expanding label %s with %s",
            label,
            strategy_generator,
        )
        if label is None:
            label = self.classdb.get_label(comb_class)

        for rule in self._rules_from_strategy(comb_class, strategy_generator):
            try:
                children = rule.children
            except StrategyDoesNotApply:
                continue
            if len(children) == 1 and rule.comb_class == children[0]:
                logger.debug(
                    "The equivalence strategy %s returned the same "
                    "combinatorial class when applied to %r",
                    str(rule).split(" ")[1],
                    comb_class,
                )
                continue
            end_labels = [self.classdb.get_label(child) for child in children]
            if rule.comb_class == comb_class:
                start_label = label
            else:
                start_label = self.classdb.get_label(rule.comb_class)

            # TODO: observe that creating this constructor could be costly,
            # e.g. Cartesian
            if self.debug:
                logger.debug(
                    "Adding combinatorial rule %s -> %s\n%s",
                    start_label,
                    tuple(end_labels),
                    rule,
                )
                try:
                    n = 4
                    for i in range(n + 1):
                        rule.sanity_check(n=i)
                    logger.debug("Sanity checked rule to length %s.", n)
                except NotImplementedError as e:
                    logger.debug(
                        "Could not sanity check rule due to:\n"
                        "NotImplementedError: %s",
                        e,
                    )
            yield start_label, tuple(end_labels), rule

    @cssmethodtimer("add rule")
    def add_rule(
        self, start_label: int, end_labels: Tuple[int, ...], rule: AbstractRule
    ) -> None:
        """
        Add the rule to the searcher

        - try to verify children combinatorial classes
        - set workability of combinatorial classes
        - symmetry expand combinatorial classes
        - add class to classqueue
        """
        for comb_class, child_label in zip(rule.children, end_labels):
            if self.symmetries and child_label not in self.symmetry_expanded:
                self._symmetry_expand(
                    comb_class, child_label
                )  # TODO: mark symmetries as empty where appropriate
            if rule.workable:
                self.classqueue.add(child_label)
            if not rule.inferrable:
                self.classqueue.set_not_inferrable(child_label)
            if not rule.possibly_empty:
                self.classdb.set_empty(child_label, empty=False)
            self.try_verify(comb_class, child_label)
        if rule.ignore_parent:
            self.classqueue.set_stop_yielding(start_label)
        self.ruledb.add(start_label, end_labels, rule)

    def _symmetry_expand(self, comb_class: CombinatorialClassType, label: int) -> None:
        """Add symmetries of combinatorial class to the database."""
        sym_labels = set([label])
        for strategy_generator in self.symmetries:
            for start_label, end_labels, rule in self._expand_class_with_strategy(
                comb_class, strategy_generator, label=label
            ):
                sym_label = end_labels[0]
                self.ruledb.add(start_label, (sym_label,), rule)
                self.classqueue.set_stop_yielding(sym_label)
                sym_labels.add(sym_label)
        self.symmetry_expanded.update(sym_labels)

    def _inferral_expand(
        self,
        comb_class: CombinatorialClassType,
        label: int,
        inferral_strategies: Tuple[CSSstrategy, ...],
        skip: Optional[CSSstrategy] = None,
    ):
        """
        Inferral expand combinatorial class with given label and inferral
        strategies.

        It will apply all inferral strategies to an combinatorial class.
        Return True if combinatorial class is inferred.
        """
        if self.debug:
            assert comb_class == self.classdb.get_class(label)
        if label in self.inferral_expanded:
            return
        self.inferral_expanded.add(label)
        for i, strategy_generator in enumerate(inferral_strategies):
            if strategy_generator == skip:
                continue
            for start_label, end_labels, rule in self._expand_class_with_strategy(
                comb_class, strategy_generator, label=label
            ):
                inf_class = rule.children[0]
                inf_label = end_labels[0]
                self.add_rule(start_label, end_labels, rule)
                self.classqueue.set_not_inferrable(start_label)
                inferral_strategies = (
                    inferral_strategies[i + 1 :] + inferral_strategies[0 : i + 1]
                )
                self._inferral_expand(
                    inf_class, inf_label, inferral_strategies, skip=strategy_generator
                )
                break
            else:
                continue
            # only breaks if inner loop breaks
            break
        self.classqueue.set_not_inferrable(label)

    def do_level(self) -> None:
        """Expand combinatorial classes in current queue. Combintorial classes
        found added to next."""
        for label, strategies, inferral in self.classqueue.do_level():
            comb_class = self.classdb.get_class(label)
            self._expand(comb_class, label, strategies, inferral)

    @cssmethodtimer("status")
    def status(self, elaborate: bool) -> str:
        """
        Return a string of the current status of the CombSpecSearcher.

        It includes:
        - number of combinatorial classes, and information about verification
        - the times spent in each of the main functions

        "elaborate" status updates are those that provide information that
        may be slow to compute
        """
        status = "CSS status:\n"
        total = sum(self.func_times.values())
        status += f"\tTotal time accounted for: {timedelta(seconds=int(total))}\n"
        status += self._css_status(total)
        status += self.classdb.status() + "\n"
        status += self.classqueue.status() + "\n"
        status += self.ruledb.status(elaborate) + "\n"
        status += self._mem_status(elaborate)
        return status

    def _css_status(self, total: float) -> str:
        table: List[Tuple[str, str, timedelta, str]] = []
        for explanation in self.func_calls:
            count = f"{self.func_calls[explanation]:,d}"
            time_spent = timedelta(seconds=int(self.func_times[explanation]))
            if total == 0:
                percentage = "? %"
            else:
                percentage = f"{int((self.func_times[explanation] * 100) / total)}%"
            table.append((explanation, count, time_spent, percentage))
        table.sort(key=lambda row: row[2], reverse=True)
        headers = ["", "Number of \napplications", "\nTime spent", "\nPercentage"]
        colalign = ("left", "right", "right", "right")
        return (
            "    "
            + tabulate.tabulate(table, headers=headers, colalign=colalign).replace(
                "\n", "\n    "
            )
            + "\n"
        )

    def _mem_status(self, elaborate: bool) -> str:
        """Provide status information related to memory usage."""

        status = "Memory Status:\n"
        table: List[Tuple[str, str]] = []
        table.append(("OS Allocated", size_to_readable(get_mem())))
        if platform.python_implementation() == "CPython":
            # Warning: "asizeof" can be very slow!
            if elaborate:
                table.append(("CSS", size_to_readable(asizeof(self))))
                table.append(("ClassDB", size_to_readable(asizeof(self.classdb))))
                table.append(("ClassQueue", size_to_readable(asizeof(self.classqueue))))
                table.append(("RuleDB", size_to_readable(asizeof(self.ruledb))))

        elif platform.python_implementation() == "PyPy":
            gc_stats = cast(Any, gc.get_stats())
            stats = [
                ("Current Memory Used", gc_stats.total_gc_memory),
                ("Current Memory Allocated", gc_stats.total_allocated_memory),
                ("Current JIT Memory Used", gc_stats.jit_backend_used),
                ("Current JIT Memory Allocated", gc_stats.jit_backend_allocated),
                ("Peak Memory Used", gc_stats.peak_memory),
                ("Peak Memory Allocated Memory Used", gc_stats.peak_allocated_memory),
            ]
            for (desc, mem) in stats:
                table.append((desc, nice_pypy_mem(mem)))
        status += "    "
        status += tabulate.tabulate(table, colalign=("left", "right")).replace(
            "\n", "\n    "
        )
        status += "\n"
        if platform.python_implementation() == "PyPy":
            gc_time = timedelta(seconds=int(gc_stats.total_gc_time / 1000))
            status += f"\tTotal Garbage Collection Time: {gc_time}\n"
        return status

    def run_information(self) -> str:
        """Return a string detailing what CombSpecSearcher is looking for."""
        start_string = (
            "Initialising CombSpecSearcher for the combinatorial"
            f" class:\n{self.start_class}\n"
        )
        start_string += str(self.strategy_pack)
        return start_string

    def _log_spec_found(
        self, specification: CombinatorialSpecification, start_time: float
    ) -> None:
        found_string = "Specification built\n"
        time_taken = time.time() - start_time
        found_string += f"Time taken: {timedelta(seconds=int(time_taken))}\n"
        found_string += self.status(elaborate=True)
        found_string += (
            f"Specification found has {specification.number_of_rules()} rules"
        )
        logger.info(found_string)

    def _log_status(self, start_time: float, status_update: int) -> None:
        time_taken = time.time() - start_time
        status = f"Time taken so far: {timedelta(seconds=int(time_taken))}\n"
        elaborate = time.time() - start_time > 100 * self.func_times["status"]
        status_start = time.time()
        status += self.status(elaborate=elaborate)
        ne_goal = 100 * self.func_times["status"] - (time.time() - start_time)
        next_elaborate = round(ne_goal - (ne_goal % status_update) + status_update)
        if elaborate:
            time_spent = time.time() - status_start
            status += f" -- status update took {time_spent:.2f} seconds --\n"
        else:
            status += (
                " -- next elaborate status update in "
                f"{timedelta(seconds=next_elaborate)} --\n"
            )
        logger.info(status)

    def auto_search(self, **kwargs) -> CombinatorialSpecification:
        """
        An automatic search function.

        Classes will be expanded until a proof tree is found. A tree will be
        searched for approximately 1% of the search time. This can be set using
        the 'perc' keyword, as some percentage between 0 and 100.

        The search will continue, unless a proof tree is found. You can set the
        keyword 'max_time' to stop the search after 'max_time' many seconds.
        If max_time is reached it will raise a ExceededMaxtimeError.

        Information is logged to logger.info. It will also log the proof tree,
        in json format. For periodic status_updates, set the keyword flag
        'status_update', an update will be given every status_update seconds.
        "Elaborate" status updates, which provide slow-to-calculate information
        will be given periodically, so that they take less than 1% of computation
        time.

        If 'smallest' is set to 'True' then the searcher will return a proof
        tree that is as small as possible.
        """
        auto_search_start = time.time()
        start_string = "Auto search started\n"
        start_string += self.run_information()
        logger.info(start_string)
        spec_rules = self._auto_search_rules(**kwargs)
        if spec_rules is not None:
            specification = CombinatorialSpecification(self.start_class, spec_rules)
            self._log_spec_found(specification, auto_search_start)
            return specification
        raise SpecificationNotFound

    def _auto_search_rules(
        self,
        *,
        max_expansion_time: Optional[float] = None,
        perc: int = 1,
        smallest: bool = False,
        status_update: Optional[int] = None,
    ) -> Iterator[AbstractRule]:
        """
        The core functionality of the auto_search method.

        This method is used by CombinatorialSpecification for expanding
        verified classes.  Will raise SpecificationNotFound error if no
        specification is found after running out of classes to expand.
        """
        if not 0 < perc <= 100:
            logger.warning(
                "Percentage not between 0 and 100, so assuming 1% search percentage."
            )
            perc = 1
        auto_search_start = time.time()
        expansion_time: float = 0
        status_start = time.time()
        expanding = True
        while expanding:
            expanding, status_start = self._expand_classes_for(
                expansion_time, status_update, status_start, auto_search_start
            )
            spec_search_start = time.time()
            logger.debug("Searching for specification.")
            if self.has_specification():
                logger.info("Specification detected.")
                return self.ruledb.get_specification_rules(
                    smallest=smallest,
                    minimization_time_limit=0.01 * (time.time() - auto_search_start),
                )
            logger.debug("No specification found.")
            if (
                max_expansion_time is not None
                and time.time() - auto_search_start > max_expansion_time
            ):
                raise ExceededMaxtimeError(
                    "Exceeded maximum time. Aborting auto search.",
                )
            # worst case, search every hour
            multiplier = 100 / perc
            expansion_time = min(multiplier * (time.time() - spec_search_start), 3600.0)
            logger.debug(
                "Will expand for %s seconds.",
                round(expansion_time, 2),
            )
        raise SpecificationNotFound

    def _expand_classes_for(
        self,
        expansion_time: float,
        status_update: Optional[int],
        status_start: float,
        auto_search_start: float,
    ) -> Tuple[bool, float]:
        """
        Will expand classes for `expansion_time` seconds.

        It will return a pair (bool, time), where the bool is True if there are
        more classes to expand, False otherwise. The `time` is the time that the
        last status update printed.
        """
        expansion_start = time.time()
        last_label = None
        expanding = True
        for label, strategies, inferral in self.classqueue:
            if label != last_label:
                comb_class = self.classdb.get_class(label)
                last_label = label
            if self.expand_verified or not self.ruledb.is_verified(label):
                self._expand(comb_class, label, strategies, inferral)
            if time.time() - expansion_start > expansion_time:
                break
            if status_update is not None and time.time() - status_start > status_update:
                self._log_status(auto_search_start, status_update)
                status_start = time.time()
        else:
            expanding = False
            logger.info("No more classes to expand.")
        return expanding, status_start

    @cssmethodtimer("has specification")
    def has_specification(self) -> bool:
        return self.ruledb.has_specification()

    @cssmethodtimer("get specification")
    def get_specification(
        self, minimization_time_limit: float = 10, smallest: bool = False
    ) -> CombinatorialSpecification:
        """
        Return a CombinatorialSpecification if the universe contains one.

        The minimization_time_limit only applies when smallest is false.

        The function will return None if no such CombinatorialSpecification
        exists in the universe.
        """
        if not self.ruledb.has_specification():
            raise SpecificationNotFound
        kwargs = {
            "minimization_time_limit": minimization_time_limit,
            "smallest": smallest,
        }
        rules = self.ruledb.get_specification_rules(**kwargs)
        logger.info("Creating a specification.")
        return CombinatorialSpecification(self.start_class, rules)
